import pytest
from mongomock import Database

from odata_v4_query.errors import (
    AggregationOperatorNotSupportedError,
    UnexpectedEmptyArgumentsError,
    UnexpectedNullFiltersError,
    UnexpectedNullFunctionNameError,
    UnexpectedNullOperandError,
    UnexpectedNullOperatorError,
    UnexpectedNumberOfArgumentsError,
    UnexpectedTypeError,
    UnknownFunctionError,
    UnknownOperatorError,
)
from odata_v4_query.filter_parser import FilterNode
from odata_v4_query.query_parser import ODataQueryOptions, ODataQueryParser
from odata_v4_query.utils.pymongo import PyMongoQuery, get_query_from_options

from ._core.pymongo import get_client, seed_data


@pytest.fixture(scope='session')
def db():
    client = get_client()
    db = client['db']
    seed_data(db)
    return db


class TestBeanie:
    parser = ODataQueryParser()

    def test_py_mongo_query_class(self):
        query = PyMongoQuery(skip=1, limit=10, filter={'name': 'John'})
        assert query.skip == 1
        assert query.limit == 10
        assert query.filter == {'name': 'John'}
        assert query['skip'] == 1
        assert query['limit'] == 10
        assert query['filter'] == {'name': 'John'}

        query.skip = 2
        assert query.skip == 2
        assert query['skip'] == 2

        query['skip'] = 3
        assert query.skip == 3
        assert query['skip'] == 3

    def test_skip(self, db: Database):
        users_count = len(list(db.users.find()))
        options = self.parser.parse_query_string('$skip=2')
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == users_count - 2
        assert result[0]['name'] == 'Alice'

    def test_top(self, db: Database):
        options = self.parser.parse_query_string('$top=2')
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == 2
        assert result[0]['name'] == 'John'
        assert result[1]['name'] == 'Jane'

    def test_page(self, db: Database):
        users_count = len(list(db.users.find()))
        options1 = self.parser.parse_query_string('$page=1')
        query1 = get_query_from_options(options1)
        result1 = list(db.users.find(**query1))
        options2 = self.parser.parse_query_string('$page=1&$top=4')
        query2 = get_query_from_options(options2)
        result2 = list(db.users.find(**query2))
        options3 = self.parser.parse_query_string('$page=2&$top=4')
        query3 = get_query_from_options(options3)
        result3 = list(db.users.find(**query3))
        options4 = self.parser.parse_query_string('$page=3&$top=4')
        query4 = get_query_from_options(options4)
        result4 = list(db.users.find(**query4))
        options5 = self.parser.parse_query_string('$page=4&$top=4')
        query5 = get_query_from_options(options5)
        result5 = list(db.users.find(**query5))
        assert len(result1) == users_count
        assert len(result2) == 4
        assert len(result3) == 4
        assert len(result4) == 2
        assert len(result5) == 0

    def test_filter_comparison(self, db: Database):
        options1 = self.parser.parse_query_string(
            "$filter=name eq 'John' and age ge 25"
        )
        query1 = get_query_from_options(options1)
        result1 = list(db.users.find(**query1))
        options2 = self.parser.parse_query_string('$filter=age lt 25 or age gt 35')
        query2 = get_query_from_options(options2)
        result2 = list(db.users.find(**query2))
        options3 = self.parser.parse_query_string("$filter=name in ('Eve', 'Frank')")
        query3 = get_query_from_options(options3)
        result3 = list(db.users.find(**query3))
        options4 = self.parser.parse_query_string("$filter=name nin ('Eve', 'Frank')")
        query4 = get_query_from_options(options4)
        result4 = list(db.users.find(**query4))
        assert len(result1) == 1
        assert result1[0]['name'] == 'John'
        assert len(result2) == 4
        assert len(result3) == 2
        assert result3[0]['name'] == 'Eve'
        assert result3[1]['name'] == 'Frank'
        assert len(result4) == 8

    def test_filter_logical(self, db: Database):
        options1 = self.parser.parse_query_string(
            "$filter=name ne 'John' and name ne 'Jane'"
        )
        query1 = get_query_from_options(options1)
        result1 = list(db.users.find(**query1))
        options2 = self.parser.parse_query_string(
            "$filter=not name eq 'John' and not name eq 'Jane'"
        )
        query2 = get_query_from_options(options2)
        result2 = list(db.users.find(**query2))
        assert len(result1) == 8
        assert len(result2) == 8

    def test_filter_null(self, db: Database):
        options1 = self.parser.parse_query_string('$filter=name eq null')
        query1 = get_query_from_options(options1)
        result1 = list(db.users.find(**query1))
        options2 = self.parser.parse_query_string('$filter=name ne null')
        query2 = get_query_from_options(options2)
        result2 = list(db.users.find(**query2))
        assert len(result1) == 0
        assert len(result2) == 10

    def test_filter_startswith_function(self, db: Database):
        options = self.parser.parse_query_string(
            "$filter=startswith(name, 'J') and age ge 25"
        )
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == 2
        assert result[0]['name'] == 'John'
        assert result[1]['name'] == 'Jane'

    def test_filter_endswith_function(self, db: Database):
        options = self.parser.parse_query_string("$filter=endswith(name, 'e')")
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == 5
        assert result[0]['name'] == 'Jane'
        assert result[1]['name'] == 'Alice'
        assert result[2]['name'] == 'Charlie'
        assert result[3]['name'] == 'Eve'
        assert result[4]['name'] == 'Grace'

    def test_filter_contains_function(self, db: Database):
        options = self.parser.parse_query_string(
            "$filter=contains(name, 'i') and age le 35"
        )
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == 2
        assert result[0]['name'] == 'Alice'
        assert result[0]['age'] == 35
        assert result[1]['name'] == 'Charlie'
        assert result[1]['age'] == 32

    def test_filter_substring_function(self, db: Database):
        options = self.parser.parse_query_string(
            "$filter=substring(name, 1, 3) eq 'ohn'"
        )
        with pytest.raises(AggregationOperatorNotSupportedError):
            get_query_from_options(options)

    def test_filter_tolower_function(self, db: Database):
        options = self.parser.parse_query_string("$filter=tolower(name) eq 'john'")
        with pytest.raises(AggregationOperatorNotSupportedError):
            get_query_from_options(options)

    def test_filter_toupper_function(self, db: Database):
        options = self.parser.parse_query_string("$filter=toupper(name) eq 'ALICE'")
        with pytest.raises(AggregationOperatorNotSupportedError):
            get_query_from_options(options)

    def test_filter_has(self, db: Database):
        options = self.parser.parse_query_string("$filter=addresses has '101 Main St'")
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == 2
        assert result[0]['name'] == 'Alice'
        assert result[1]['name'] == 'Bob'

    def test_search(self, db: Database):
        options = self.parser.parse_query_string('$search=John')
        query = get_query_from_options(options, search_fields=['name', 'email'])
        result = list(db.users.find(**query))
        assert len(result) == 1
        assert result[0]['name'] == 'John'

    def test_filter_and_search_merge(self, db: Database):
        options = self.parser.parse_query_string('$filter=age gt 30&$search=a')
        query = get_query_from_options(options, search_fields=['name'])
        result = list(db.users.find(**query))
        # Should find users with age > 30 AND name containing 'a' (case-sensitive)
        assert len(result) == 3
        names = [r['name'] for r in result]
        assert 'Charlie' in names  # age 32, name contains 'a'
        assert 'David' in names  # age 41, name contains 'a'
        assert 'Grace' in names  # age 37, name contains 'a'

    def test_orderby(self, db: Database):
        options = self.parser.parse_query_string('$orderby=name asc,age desc')
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == 10
        assert result[0]['name'] == 'Alice'
        assert result[1]['name'] == 'Bob'
        assert result[1]['age'] == 40
        assert result[2]['name'] == 'Bob'
        assert result[2]['age'] == 28
        assert result[3]['name'] == 'Charlie'
        assert result[4]['name'] == 'David'
        assert result[5]['name'] == 'Eve'
        assert result[6]['name'] == 'Frank'
        assert result[7]['name'] == 'Grace'
        assert result[8]['name'] == 'Jane'
        assert result[9]['name'] == 'John'

    def test_select(self, db: Database):
        options = self.parser.parse_query_string('$select=name,email')
        query = get_query_from_options(options, parse_select=True)
        result = list(db.users.find(**query))
        assert len(result) == 10
        assert result[0]['name'] == 'John'
        assert result[0]['email'] == 'john@example.com'

    def test_filter_nested_field_eq(self, db: Database):
        options = self.parser.parse_query_string("$filter=profile/city eq 'Chicago'")
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == 1
        assert result[0]['name'] == 'Alice'
        assert result[0]['profile']['city'] == 'Chicago'

    def test_filter_nested_field_comparison(self, db: Database):
        options = self.parser.parse_query_string(
            "$filter=profile/city eq 'New York' and age gt 20"
        )
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == 1
        assert result[0]['name'] == 'John'

    def test_filter_nested_field_startswith(self, db: Database):
        options = self.parser.parse_query_string(
            "$filter=startswith(profile/city, 'San')"
        )
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == 3
        names = [r['name'] for r in result]
        assert 'David' in names  # San Antonio
        assert 'Eve' in names  # San Diego
        assert 'Grace' in names  # San Jose

    def test_filter_nested_field_contains(self, db: Database):
        options = self.parser.parse_query_string(
            "$filter=contains(profile/city, 'Angeles')"
        )
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == 1
        assert result[0]['name'] == 'Jane'

    def test_filter_nested_field_in_operator(self, db: Database):
        options = self.parser.parse_query_string(
            "$filter=profile/city in ('Chicago', 'Houston', 'Dallas')"
        )
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == 3
        names = [r['name'] for r in result]
        assert 'Alice' in names  # Chicago
        assert 'Bob' in names  # Houston
        assert 'Frank' in names  # Dallas

    def test_filter_nested_field_has_operator(self, db: Database):
        options = self.parser.parse_query_string("$filter=profile/country has 'USA'")
        query = get_query_from_options(options)
        result = list(db.users.find(**query))
        assert len(result) == 10

    def test_unexpected_null_filters(self):
        options = ODataQueryOptions(filter_=FilterNode(type_='value'))
        with pytest.raises(UnexpectedNullFiltersError):
            get_query_from_options(options)

    def test_unexpected_null_operator(self):
        options = ODataQueryOptions(filter_=FilterNode(type_='operator'))
        with pytest.raises(UnexpectedNullOperatorError):
            get_query_from_options(options)

    def test_unexpected_null_operand(self):
        options = ODataQueryOptions(filter_=FilterNode(type_='operator', value='eq'))
        with pytest.raises(UnexpectedNullOperandError):
            get_query_from_options(options)

    def test_unexpected_null_operand_value(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='operator',
                value='eq',
                left=FilterNode(type_='identifier'),
                right=FilterNode(type_='literal', value='John'),
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            get_query_from_options(options)

    def test_unexpected_null_operand_for_in_nin_operators(self):
        options1 = ODataQueryOptions(
            filter_=FilterNode(
                type_='operator',
                value='in',
                left=FilterNode(type_='identifier', value='name'),
                right=FilterNode(type_='list'),
            )
        )
        options2 = ODataQueryOptions(
            filter_=FilterNode(
                type_='operator',
                value='nin',
                left=FilterNode(type_='identifier', value='name'),
                right=FilterNode(type_='list'),
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            get_query_from_options(options1)
        with pytest.raises(UnexpectedNullOperandError):
            get_query_from_options(options2)

    def test_unexpected_null_operand_for_has_operator(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='operator',
                value='has',
                left=FilterNode(type_='identifier', value='addresses'),
                right=FilterNode(type_='literal'),
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            get_query_from_options(options)

    def test_unexpected_null_operand_for_and_or_operators(self):
        options1 = ODataQueryOptions(filter_=FilterNode(type_='operator', value='and'))
        options2 = ODataQueryOptions(filter_=FilterNode(type_='operator', value='or'))
        with pytest.raises(UnexpectedNullOperandError):
            get_query_from_options(options1)
        with pytest.raises(UnexpectedNullOperandError):
            get_query_from_options(options2)

    def test_unexpected_null_operand_for_not_nor_operators(self):
        options1 = ODataQueryOptions(filter_=FilterNode(type_='operator', value='not'))
        options2 = ODataQueryOptions(filter_=FilterNode(type_='operator', value='nor'))
        with pytest.raises(UnexpectedNullOperandError):
            get_query_from_options(options1)
        with pytest.raises(UnexpectedNullOperandError):
            get_query_from_options(options2)

    def test_unknown_operator(self):
        options = ODataQueryOptions(
            filter_=FilterNode(type_='operator', value='unknown')
        )
        with pytest.raises(UnknownOperatorError):
            get_query_from_options(options)

    def test_unexpected_null_function_name(self):
        options = ODataQueryOptions(filter_=FilterNode(type_='function'))
        with pytest.raises(UnexpectedNullFunctionNameError):
            get_query_from_options(options)

    def test_unexpected_empty_arguments(self):
        options = ODataQueryOptions(
            filter_=FilterNode(type_='function', value='startswith')
        )
        with pytest.raises(UnexpectedEmptyArgumentsError):
            get_query_from_options(options)

    def test_unexpected_number_of_arguments(self):
        options1 = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='startswith',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal', value='J'),
                    FilterNode(type_='literal', value='J'),
                ],
            )
        )
        options2 = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='substring',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal', value=1),
                ],
            )
        )
        with pytest.raises(UnexpectedNumberOfArgumentsError):
            get_query_from_options(options1)
        with pytest.raises(UnexpectedNumberOfArgumentsError):
            get_query_from_options(options2)

    def test_unexpected_null_operand_for_function(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='startswith',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal'),
                ],
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            get_query_from_options(options)

    def test_unknown_function(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='unknown',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal', value='J'),
                ],
            )
        )
        with pytest.raises(UnknownFunctionError):
            get_query_from_options(options)

    def test_function_with_null_field(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='startswith',
                arguments=[
                    FilterNode(type_='identifier'),
                    FilterNode(type_='literal', value='J'),
                ],
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            get_query_from_options(options)

    def test_function_with_invalid_type(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='substring',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal', value='invalid'),
                    FilterNode(type_='literal', value=2),
                ],
            )
        )
        with pytest.raises(UnexpectedTypeError):
            get_query_from_options(options)



import pytest
import pytest_asyncio

from odata_v4_query.errors import (
    AggregationOperatorNotSupportedError,
    UnexpectedEmptyArgumentsError,
    UnexpectedNullFiltersError,
    UnexpectedNullFunctionNameError,
    UnexpectedNullOperandError,
    UnexpectedNullOperatorError,
    UnexpectedNumberOfArgumentsError,
    UnexpectedTypeError,
    UnknownFunctionError,
    UnknownOperatorError,
)
from odata_v4_query.filter_parser import FilterNode
from odata_v4_query.query_parser import ODataQueryOptions, ODataQueryParser
from odata_v4_query.utils.beanie import apply_to_beanie_query

from ._core.beanie import User, UserProjection, get_client, seed_data

"""
Monkey patching AggregationQuery to add a modified get_cursor
that does not await the ``aggregate`` call. This is a workaround
for the ``TypeError: object AsyncIOMotorLatentCommandCursor can't be used in 'await' expression``
error raised with beanie>1.23.0 and mongomock-motor>0.0.35.
"""
try:
    from beanie.odm.queries.aggregation import AggregationQuery

    async def get_cursor(self):
        aggregation_pipeline = self.get_aggregation_pipeline()
        return self.document_model.get_pymongo_collection().aggregate(
            aggregation_pipeline, session=self.session, **self.pymongo_kwargs
        )

    AggregationQuery.get_cursor = get_cursor
except ImportError as e:
    pass


@pytest_asyncio.fixture(autouse=True)
async def client():
    client = await get_client()
    await seed_data()
    return client


class TestBeanie:
    parser = ODataQueryParser()

    @pytest.mark.asyncio
    async def test_skip(self):
        users_count = len(await User.find().to_list())
        options = self.parser.parse_query_string('$skip=2')
        query = apply_to_beanie_query(options, User.find())
        result = await query.to_list()
        assert len(result) == users_count - 2
        assert result[0].name == 'Alice'

    @pytest.mark.asyncio
    async def test_top(self):
        options = self.parser.parse_query_string('$top=2')
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 2
        assert result[0].name == 'John'
        assert result[1].name == 'Jane'

    @pytest.mark.asyncio
    async def test_page(self):
        users_count = len(await User.find().to_list())

        # default top
        options = self.parser.parse_query_string('$page=1')
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == users_count

        # top 3
        options = self.parser.parse_query_string('$page=1&$top=4')
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 4
        options = self.parser.parse_query_string('$page=2&$top=4')
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 4
        options = self.parser.parse_query_string('$page=3&$top=4')
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 2
        options = self.parser.parse_query_string('$page=4&$top=4')
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 0

    @pytest.mark.asyncio
    async def test_filter_comparison_and_logical(self):
        options = self.parser.parse_query_string("$filter=name eq 'John' and age ge 25")
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 1
        assert result[0].name == 'John'

        options = self.parser.parse_query_string('$filter=age lt 25 or age gt 35')
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 4

        options = self.parser.parse_query_string("$filter=name in ('Eve', 'Frank')")
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 2
        assert result[0].name == 'Eve'
        assert result[1].name == 'Frank'

        options = self.parser.parse_query_string("$filter=name nin ('Eve', 'Frank')")
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 8

        options = self.parser.parse_query_string(
            "$filter=name ne 'John' and name ne 'Jane'"
        )
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 8

        options = self.parser.parse_query_string(
            "$filter=not name eq 'John' and not name eq 'Jane'"
        )
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 8

        options = self.parser.parse_query_string('$filter=name eq null')
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 0

        options = self.parser.parse_query_string('$filter=name ne null')
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 10

    @pytest.mark.asyncio
    async def test_filter_startswith_function(self):
        options = self.parser.parse_query_string(
            "$filter=startswith(name, 'J') and age ge 25"
        )
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 2
        assert result[0].name == 'John'
        assert result[1].name == 'Jane'

    @pytest.mark.asyncio
    async def test_filter_endswith_function(self):
        options = self.parser.parse_query_string("$filter=endswith(name, 'e')")
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 5
        assert result[0].name == 'Jane'
        assert result[1].name == 'Alice'
        assert result[2].name == 'Charlie'
        assert result[3].name == 'Eve'
        assert result[4].name == 'Grace'

    @pytest.mark.asyncio
    async def test_filter_contains_function(self):
        options = self.parser.parse_query_string(
            "$filter=contains(name, 'i') and age le 35"
        )
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 2
        assert result[0].name == 'Alice'
        assert result[0].age == 35
        assert result[1].name == 'Charlie'
        assert result[1].age == 32

    @pytest.mark.asyncio
    async def test_filter_substring_function(self):
        options = self.parser.parse_query_string(
            "$filter=substring(name, 1, 3) eq 'ohn'"
        )
        with pytest.raises(AggregationOperatorNotSupportedError):
            apply_to_beanie_query(options, User)

    @pytest.mark.asyncio
    async def test_filter_tolower_function(self):
        options = self.parser.parse_query_string("$filter=tolower(name) eq 'john'")
        with pytest.raises(AggregationOperatorNotSupportedError):
            apply_to_beanie_query(options, User)

    @pytest.mark.asyncio
    async def test_filter_toupper_function(self):
        options = self.parser.parse_query_string("$filter=toupper(name) eq 'ALICE'")
        with pytest.raises(AggregationOperatorNotSupportedError):
            apply_to_beanie_query(options, User)

    @pytest.mark.asyncio
    async def test_filter_has_operator(self):
        options = self.parser.parse_query_string("$filter=addresses has '101 Main St'")
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 2
        assert result[0].name == 'Alice'
        assert result[1].name == 'Bob'

    @pytest.mark.asyncio
    async def test_search(self):
        options = self.parser.parse_query_string('$search=John')
        query = apply_to_beanie_query(options, User, search_fields=['name', 'email'])
        result = await query.to_list()
        assert len(result) == 1
        assert result[0].name == 'John'

    @pytest.mark.asyncio
    async def test_count(self):
        options = self.parser.parse_query_string('$count=true')
        result = apply_to_beanie_query(options, User, count=True)
        count = await result
        assert count == 10

    @pytest.mark.asyncio
    async def test_orderby(self):
        options = self.parser.parse_query_string('$orderby=name asc,age desc')
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 10
        assert result[0].name == 'Alice'
        assert result[1].name == 'Bob'
        assert result[1].age == 40
        assert result[2].name == 'Bob'
        assert result[2].age == 28
        assert result[3].name == 'Charlie'
        assert result[4].name == 'David'
        assert result[5].name == 'Eve'
        assert result[6].name == 'Frank'
        assert result[7].name == 'Grace'
        assert result[8].name == 'Jane'
        assert result[9].name == 'John'

    @pytest.mark.asyncio
    async def test_select(self):
        options = self.parser.parse_query_string('$select=name,email')
        query = apply_to_beanie_query(options, User, parse_select=True)
        result = await query.to_list()
        assert len(result) == 10
        assert result[0]['name'] == 'John'
        assert result[0]['email'] == 'john@example.com'

    @pytest.mark.asyncio
    async def test_filter_nested_field_eq(self):
        options = self.parser.parse_query_string("$filter=profile/city eq 'Chicago'")
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 1
        assert result[0].name == 'Alice'
        assert result[0].profile.city == 'Chicago'

    @pytest.mark.asyncio
    async def test_filter_nested_field_comparison(self):
        options = self.parser.parse_query_string(
            "$filter=profile/city eq 'New York' and age gt 20"
        )
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 1
        assert result[0].name == 'John'

    @pytest.mark.asyncio
    async def test_filter_nested_field_startswith(self):
        options = self.parser.parse_query_string(
            "$filter=startswith(profile/city, 'San')"
        )
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 3
        names = [r.name for r in result]
        assert 'David' in names  # San Antonio
        assert 'Eve' in names  # San Diego
        assert 'Grace' in names  # San Jose

    @pytest.mark.asyncio
    async def test_filter_nested_field_contains(self):
        options = self.parser.parse_query_string(
            "$filter=contains(profile/city, 'Angeles')"
        )
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 1
        assert result[0].name == 'Jane'

    @pytest.mark.asyncio
    async def test_filter_nested_field_in_operator(self):
        options = self.parser.parse_query_string(
            "$filter=profile/city in ('Chicago', 'Houston', 'Dallas')"
        )
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 3
        names = [r.name for r in result]
        assert 'Alice' in names  # Chicago
        assert 'Bob' in names  # Houston
        assert 'Frank' in names  # Dallas

    @pytest.mark.asyncio
    async def test_filter_nested_field_has_operator(self):
        options = self.parser.parse_query_string("$filter=profile/country has 'USA'")
        query = apply_to_beanie_query(options, User)
        result = await query.to_list()
        assert len(result) == 10

    @pytest.mark.asyncio
    async def test_projection(self):
        options = self.parser.parse_query_string('$top=1')
        query = apply_to_beanie_query(options, User, projection_model=UserProjection)
        result = await query.to_list()
        assert len(result) == 1
        assert isinstance(result[0], UserProjection)
        assert result[0].name == 'John'
        assert result[0].email == 'john@example.com'

        options = self.parser.parse_query_string('$top=1&$select=name,email')
        query = apply_to_beanie_query(
            options, User, projection_model=UserProjection, parse_select=True
        )
        result = await query.to_list()
        assert len(result) == 1
        assert isinstance(result[0], UserProjection)
        assert result[0].name == 'John'
        assert result[0].email == 'john@example.com'

    def test_unexpected_null_filters(self):
        options = ODataQueryOptions(filter_=FilterNode(type_='value'))
        with pytest.raises(UnexpectedNullFiltersError):
            apply_to_beanie_query(options, User)

    def test_unexpected_null_operator(self):
        options = ODataQueryOptions(filter_=FilterNode(type_='operator'))
        with pytest.raises(UnexpectedNullOperatorError):
            apply_to_beanie_query(options, User)

    def test_unexpected_null_operand(self):
        options = ODataQueryOptions(filter_=FilterNode(type_='operator', value='eq'))
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_beanie_query(options, User)

    def test_unexpected_null_operand_value(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='operator',
                value='eq',
                left=FilterNode(type_='identifier'),
                right=FilterNode(type_='literal', value='John'),
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_beanie_query(options, User)

    def test_unexpected_null_operand_for_in_nin_operators(self):
        options1 = ODataQueryOptions(
            filter_=FilterNode(
                type_='operator',
                value='in',
                left=FilterNode(type_='identifier', value='name'),
                right=FilterNode(type_='list'),
            )
        )
        options2 = ODataQueryOptions(
            filter_=FilterNode(
                type_='operator',
                value='nin',
                left=FilterNode(type_='identifier', value='name'),
                right=FilterNode(type_='list'),
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_beanie_query(options1, User)
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_beanie_query(options2, User)

    def test_unexpected_null_operand_for_has_operator(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='operator',
                value='has',
                left=FilterNode(type_='identifier', value='addresses'),
                right=FilterNode(type_='literal'),
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_beanie_query(options, User)

    def test_unexpected_null_operand_for_and_or_operators(self):
        options1 = ODataQueryOptions(filter_=FilterNode(type_='operator', value='and'))
        options2 = ODataQueryOptions(filter_=FilterNode(type_='operator', value='or'))
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_beanie_query(options1, User)
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_beanie_query(options2, User)

    def test_unexpected_null_operand_for_not_nor_operators(self):
        options1 = ODataQueryOptions(filter_=FilterNode(type_='operator', value='not'))
        options2 = ODataQueryOptions(filter_=FilterNode(type_='operator', value='nor'))
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_beanie_query(options1, User)
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_beanie_query(options2, User)

    def test_unknown_operator(self):
        options = ODataQueryOptions(
            filter_=FilterNode(type_='operator', value='unknown')
        )
        with pytest.raises(UnknownOperatorError):
            apply_to_beanie_query(options, User)

    def test_unexpected_null_function_name(self):
        options = ODataQueryOptions(filter_=FilterNode(type_='function'))
        with pytest.raises(UnexpectedNullFunctionNameError):
            apply_to_beanie_query(options, User)

    def test_unexpected_empty_arguments(self):
        options = ODataQueryOptions(
            filter_=FilterNode(type_='function', value='startswith')
        )
        with pytest.raises(UnexpectedEmptyArgumentsError):
            apply_to_beanie_query(options, User)

    def test_unexpected_number_of_arguments(self):
        options1 = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='startswith',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal', value='J'),
                    FilterNode(type_='literal', value='J'),
                ],
            )
        )
        options2 = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='substring',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal', value=1),
                ],
            )
        )
        with pytest.raises(UnexpectedNumberOfArgumentsError):
            apply_to_beanie_query(options1, User)
        with pytest.raises(UnexpectedNumberOfArgumentsError):
            apply_to_beanie_query(options2, User)

    def test_unexpected_null_operand_for_function(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='startswith',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal'),
                ],
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_beanie_query(options, User)

    def test_unknown_function(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='unknown',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal', value='J'),
                ],
            )
        )
        with pytest.raises(UnknownFunctionError):
            apply_to_beanie_query(options, User)

    def test_function_with_null_field(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='startswith',
                arguments=[
                    FilterNode(type_='identifier'),
                    FilterNode(type_='literal', value='J'),
                ],
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_beanie_query(options, User)

    def test_function_with_invalid_type(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='substring',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal', value='invalid'),
                    FilterNode(type_='literal', value=2),
                ],
            )
        )
        with pytest.raises(UnexpectedTypeError):
            apply_to_beanie_query(options, User)



"""Utility function for applying OData query options to a Beanie query.

See ``apply_to_beanie_query()`` for more information.
"""

try:
    from beanie import Document
    from beanie.odm.queries.aggregation import AggregationQuery
    from beanie.odm.queries.find import FindMany
    from beanie.operators import Or
except ImportError as e:  # pragma: no cover
    missing_dep_msg = (
        'The beanie dependency is not installed. '
        'Install it with `pip install odata-v4-query[beanie]` '
        'or install it directly with `pip install beanie`.'
    )
    raise ImportError(missing_dep_msg) from e  # pragma: no cover

from collections.abc import Coroutine
from typing import Any, Literal, TypeVar, overload

from pydantic import BaseModel

from odata_v4_query.query_parser import ODataQueryOptions

from ._func import compute_skip_from_page
from .filter_parsers.mongo_filter_parser import MongoDBFilterNodeParser

FindQueryProjectionType = TypeVar('FindQueryProjectionType', bound=BaseModel)
FindType = TypeVar('FindType', bound=Document)
ParsedQuery = (
    FindMany[FindType]
    | FindMany[FindQueryProjectionType]
    | AggregationQuery[dict[str, Any]]
    | AggregationQuery[FindQueryProjectionType]
    | Coroutine[Any, Any, int]
)


@overload
def apply_to_beanie_query(
    options: ODataQueryOptions,
    document_or_query: type[FindType] | FindMany[FindType],
    projection_model: None = None,
    parse_select: Literal[False] = False,
    search_fields: list[str] | None = None,
    fetch_links: bool = False,
    count: Literal[False] = False,
) -> FindMany[FindType]: ...


@overload
def apply_to_beanie_query(
    options: ODataQueryOptions,
    document_or_query: type[FindType] | FindMany[FindType],
    projection_model: type[FindQueryProjectionType],
    parse_select: Literal[False] = False,
    search_fields: list[str] | None = None,
    fetch_links: bool = False,
    count: Literal[False] = False,
) -> FindMany[FindQueryProjectionType]: ...


@overload
def apply_to_beanie_query(
    options: ODataQueryOptions,
    document_or_query: type[FindType] | FindMany[FindType],
    projection_model: None = None,
    parse_select: Literal[True] = True,
    search_fields: list[str] | None = None,
    fetch_links: bool = False,
    count: Literal[False] = False,
) -> AggregationQuery[dict[str, Any]]: ...


@overload
def apply_to_beanie_query(
    options: ODataQueryOptions,
    document_or_query: type[FindType] | FindMany[FindType],
    projection_model: type[FindQueryProjectionType],
    parse_select: Literal[True] = True,
    search_fields: list[str] | None = None,
    fetch_links: bool = False,
    count: Literal[False] = False,
) -> AggregationQuery[FindQueryProjectionType]: ...


@overload
def apply_to_beanie_query(
    options: ODataQueryOptions,
    document_or_query: type[FindType] | FindMany[FindType],
    projection_model: type[FindQueryProjectionType] | None = None,
    parse_select: bool = False,
    search_fields: list[str] | None = None,
    fetch_links: bool = False,
    count: Literal[True] = True,
) -> Coroutine[Any, Any, int]: ...


def apply_to_beanie_query(
    options: ODataQueryOptions,
    document_or_query: type[FindType] | FindMany[FindType],
    projection_model: type[FindQueryProjectionType] | None = None,
    parse_select: bool = False,
    search_fields: list[str] | None = None,
    fetch_links: bool = False,
    count: bool = False,
) -> ParsedQuery:
    """Apply OData query options to a Beanie query.

    Parameters
    ----------
    options : ODataQueryOptions
        Parsed query options.
    document_or_query : type[FindType] | FindMany[FindType]
        Document class or query to apply options to.
        If a class is provided, a new query is created
        by calling ``find()``.
    projection_model : type[FindQueryProjectionType] | None, optional
        Projection model, by default None.
    parse_select : bool, optional
        If True, ``$select`` is parsed and applied as a projection,
        by default False.
    search_fields : list[str] | None, optional
        Fields to search in if ``$search`` is used, by default None.
    fetch_links : bool, optional
        Whether to fetch links, by default False.
    count : bool, optional
        Whether to return a coroutine that returns the count of the
        query, by default False.

    Returns
    -------
    ParsedQuery
        Beanie query with applied options.

    Notes
    -----
    #### Pagination
    If ``$page`` option is provided, ``$skip`` is overwritten
    using ``odata_v4_query.utils.compute_skip_from_page()``.
    If ``$top`` is not provided, it defaults to
    ``odata_v4_query.definitions.DEFAULT_LIMIT``.

    #### Search
    The ``$search`` option is only supported if ``search_fields``
    is provided.

    #### Select and projection
    The ``$select`` option is only supported if ``parse_select``
    is True. If ``projection_model`` is provided, the results
    are projected with a Pydantic model, otherwise a dictionary.

    #### Count
    If ``$count`` is provided, the query is returned with a
    coroutine that returns the count of the query. Also, ``$orderby``
    and ``$select`` are ignored.

    #### Unsupported options
    The ``$expand`` and ``$format`` options are not supported.

    #### Unsupported functions
    The following functions are not supported:
    - ``substring``
    - ``tolower``
    - ``toupper``

    Examples
    --------
    Assuming the following ``User`` beanie document:
    >>> from beanie import Document
    >>> class User(Document):
    ...     name: str
    ...     email: str
    ...     age: int

    Create a new query with applied options:
    >>> from odata_v4_query import ODataQueryParser
    >>> from odata_v4_query.utils.beanie import apply_to_beanie_query
    >>> parser = ODataQueryParser()
    >>> options = parser.parse_query_string('$top=10&$skip=20')
    >>> query = apply_to_beanie_query(options, User)

    Apply options to an existing query:
    >>> query = apply_to_beanie_query(options, User.find())

    Apply options with projection:
    >>> from pydantic import BaseModel
    >>> class UserProjection(BaseModel):
    ...     name: str
    ...     email: str
    >>> query = apply_to_beanie_query(
    ...     options,
    ...     User.find(),
    ...     projection_model=UserProjection
    ... )

    Apply options with projection and parsing ``$select``:
    >>> query = apply_to_beanie_query(
    ...     options,
    ...     User.find(),
    ...     projection_model=UserProjection,
    ...     parse_select=True
    ... )

    Apply ``$search`` option:
    >>> query = apply_to_beanie_query(
    ...     options,
    ...     User.find(),
    ...     search_fields=['name', 'email']
    ... )

    Fetch links:
    >>> query = apply_to_beanie_query(
    ...     options,
    ...     User.find(),
    ...     fetch_links=True
    ... )

    Get count:
    >>> count = await apply_to_beanie_query(
    ...     options,
    ...     User.find(),
    ...     count=True
    ... )()

    """
    compute_skip_from_page(options)

    if isinstance(document_or_query, FindMany):
        query = document_or_query
    else:
        query = document_or_query.find()

    if options.skip:
        query = query.skip(options.skip)

    if options.top:
        query = query.limit(options.top)

    if options.filter_:
        parser = MongoDBFilterNodeParser()
        filters = parser.parse(options.filter_)
        query = query.find(filters, fetch_links=fetch_links)

    if options.search and search_fields:
        query = query.find(
            Or(
                *[{field: {'$regex': options.search}} for field in search_fields],
            ).query,
            fetch_links=fetch_links,
        )

    if options.count and count:
        return query.count()

    if options.orderby:
        sort_args = []
        for item in options.orderby:
            direction = '-' if item.direction == 'desc' else '+'
            sort_args.append(f'{direction}{item.field}')
        query = query.sort(*sort_args)

    if options.select and parse_select:
        return query.aggregate(
            [{'$project': dict.fromkeys(options.select, 1)}],
            projection_model=projection_model,
        )

    return query.project(projection_model)



"""Utility function for applying OData query options to a SQLAlchemy query.

See ``apply_to_sqlalchemy_query()`` for more information.
"""

try:
    from sqlalchemy.orm import joinedload
    from sqlalchemy.sql import Select, or_, select
    from sqlalchemy.sql.functions import func
except ImportError as e:  # pragma: no cover
    missing_dep_msg = (
        'The sqlalchemy dependency is not installed. '
        'Install it with `pip install odata-v4-query[sqlalchemy]` '
        'or install it directly with `pip install sqlalchemy`.'
    )
    raise ImportError(missing_dep_msg) from e  # pragma: no cover

from typing import Literal, TypeVar, overload

from odata_v4_query.errors import NoRootClassError
from odata_v4_query.query_parser import ODataQueryOptions

from ._func import compute_skip_from_page
from .filter_parsers.sql_filter_parser import (
    FilterType,
    SQLAlchemyFilterNodeParser,
)

T = TypeVar('T')
Query = Select[tuple[T, ...]]
ParsedQuery = Select[tuple[FilterType, ...]] | Select


@overload
def get_query_root_cls(
    query: Query[T],
    raise_on_none: Literal[True],
) -> type[T]: ...


@overload
def get_query_root_cls(
    query: Query[T],
    raise_on_none: bool = False,
) -> type[T] | None: ...


def get_query_root_cls(
    query: Query[T],
    raise_on_none: bool = False,
) -> type[T] | None:
    """Return the root class of a query.

    Returns the class of the first column of the query, this is:
    - When selecting specific columns, returns the class of
        the first column (the same for functions like ``count()``).
    - When selecting multiple classes, returns the class of
        the first class.
    - When joining, returns the class of the first table.

    Parameters
    ----------
    query : Query[T]
        SQLAlchemy query.
    raise_on_none : bool, optional
        If True, raises an error if no root class is found,
        by default False.

    Returns
    -------
    type[T] | None
        Root class of the query.

    Raises
    ------
    NoRootClassError
        If no root class is found and ``raise_on_none`` is True.

    Examples
    --------
    Usage:
    >>> query = select(User)
    >>> get_query_root_cls(query)
    <class 'User'>

    Selecting specific columns:
    >>> query = select(User.id)
    >>> get_query_root_cls(query)
    <class 'User'>

    Selecting specific columns with functions:
    >>> query = select(func.count(User.id))
    >>> get_query_root_cls(query)
    <class 'User'>

    Selecting multiple classes:
    >>> query = select(User, Post)
    >>> get_query_root_cls(query)
    <class 'User'>

    Selecting specific columns from multiple classes:
    >>> query = select(User.id, Post.title)
    >>> get_query_root_cls(query)
    <class 'User'>

    Joining:
    >>> query = select(User).join(Post)
    >>> get_query_root_cls(query)
    <class 'User'>
    >>> query = select(Post).join(User)
    >>> get_query_root_cls(query)
    <class 'Post'>

    No root class found:
    >>> query = select(func.count('*'))
    >>> get_query_root_cls(query)
    None
    >>> get_query_root_cls(query, raise_on_none=True)
    Traceback (most recent call last):
        ...
    NoRootClassError: could not find root class of query: <...>

    """
    for col_desc in query.column_descriptions:
        entity = col_desc.get('entity')
        if entity is not None:
            return entity

    if raise_on_none:
        raise NoRootClassError(str(query))

    return None


@overload
def apply_to_sqlalchemy_query(
    options: ODataQueryOptions,
    model_or_query: type[FilterType],
    search_fields: list[str] | None = None,
) -> Select[tuple[FilterType, ...]]: ...


@overload
def apply_to_sqlalchemy_query(
    options: ODataQueryOptions,
    model_or_query: Select,
    search_fields: list[str] | None = None,
) -> Select: ...


def apply_to_sqlalchemy_query(  # noqa: C901, PLR0912
    options: ODataQueryOptions,
    model_or_query: type[FilterType] | Select,
    search_fields: list[str] | None = None,
) -> ParsedQuery:
    """Apply OData query options to a SQLAlchemy query.

    Parameters
    ----------
    options : ODataQueryOptions
        Parsed query options.
    model_or_query : type[FilterType] | Select
        SQLAlchemy model class or query to apply options to.
    search_fields : list[str] | None, optional
        Fields to search in if ``$search`` is used, by default None.

    Returns
    -------
    ParsedQuery
        SQLAlchemy query with applied options.

    Raises
    ------
    NoRootClassFound
        If the root class of the query cannot be found.

    Notes
    -----
    #### Pagination
    If ``$page`` option is provided, ``$skip`` is overwritten
    using ``odata_v4_query.utils.compute_skip_from_page()``.
    If ``$top`` is not provided, it defaults to
    ``odata_v4_query.definitions.DEFAULT_LIMIT``.

    #### Search
    The ``$search`` option is only supported if ``search_fields``
    is provided.

    #### Expand
    The ``$expand`` option performs a
    `joined eager loading <https://docs.sqlalchemy.org/en/14/orm/loading_relationships.html#sqlalchemy.orm.joinedload>`_
    using left outer join.

    #### Count
    If ``$count`` is provided, the query is returned with a
    ``count(*)`` column. The columns of the original query
    are discarded. Also, ``$orderby``, ``$expand`` and ``$select``
    are ignored.

    #### Unsupported options
    The ``$format`` option is not supported.

    Examples
    --------
    Assuming the following ``User`` SQLAlchemy model:
    >>> from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column
    >>> class User(DeclarativeBase):
    ...     name: Mapped[str] = mapped_column()
    ...     email: Mapped[str] = mapped_column()
    ...     age: Mapped[int] = mapped_column()

    Create a new query with applied options:
    >>> from odata_v4_query import ODataQueryParser
    >>> from odata_v4_query.utils.sqlalchemy import apply_to_sqlalchemy_query
    >>> parser = ODataQueryParser()
    >>> options = parser.parse_query_string('$top=10&$skip=20')
    >>> query = apply_to_sqlalchemy_query(options, User)

    Apply options to an existing query:
    >>> query = select(User)
    >>> query = apply_to_sqlalchemy_query(options, query)

    Apply ``$search`` option:
    >>> query = apply_to_sqlalchemy_query(
    ...     options,
    ...     User,
    ...     search_fields=['name', 'email']
    ... )
    >>> query = apply_to_sqlalchemy_query(
    ...     options,
    ...     select(User),
    ...     search_fields=['name', 'email']
    ... )

    """
    compute_skip_from_page(options)

    if isinstance(model_or_query, Select):
        query = model_or_query
        root_cls = get_query_root_cls(query)
    else:
        query = select(model_or_query)
        root_cls = model_or_query

    if options.skip:
        query = query.offset(options.skip)

    if options.top:
        query = query.limit(options.top)

    if options.filter_:
        if not root_cls:
            raise NoRootClassError(str(query), '$filter')

        parser = SQLAlchemyFilterNodeParser(root_cls)
        filters = parser.parse(options.filter_)
        query = query.where(filters)

    if options.search and search_fields:
        if not root_cls:
            raise NoRootClassError(str(query), '$search')

        query = query.where(
            or_(
                *[
                    getattr(root_cls, field).ilike(f'%{options.search}%')
                    for field in search_fields
                ],
            ),
        )

    if options.count:
        return query.with_only_columns(func.count('*'), maintain_column_froms=True)

    if options.orderby:
        if not root_cls:
            raise NoRootClassError(str(query), '$orderby')

        for item in options.orderby:
            if item.direction == 'asc':
                query = query.order_by(getattr(root_cls, item.field))
            else:
                query = query.order_by(getattr(root_cls, item.field).desc())

    if options.expand:
        if not root_cls:
            raise NoRootClassError(str(query), '$expand')

        for field in options.expand:
            query = query.options(joinedload(getattr(root_cls, field)))

    if options.select:
        if not root_cls:
            raise NoRootClassError(str(query), '$select')

        query = query.with_only_columns(
            *[getattr(root_cls, field) for field in options.select],
        )

    return query



"""Utility functions for applying OData query options to ORM/ODM queries.

You to need to install the required dependencies for the
ORM/ODM you want to use.

**Dependencies**

- ``beanie`` for Beanie ODM utils.
- ``pymongo`` for PyMongo ODM utils.
- ``sqlalchemy`` for SQLAlchemy ORM utils.

**Example usage**

>>> from odata_v4_query import ODataQueryParser
>>> from odata_v4_query.utils.beanie import apply_to_beanie_query
>>> # Create parser instance
>>> parser = ODataQuery_parser()
>>> # Parse a complete URL
>>> options = parser.parse_url('https://example.com/odata?$count=true&$top=10&$skip=20')
>>> # Apply options to Beanie query
>>> query = apply_to_beanie_query(options, User)
"""

from ._func import compute_skip_from_page, remove_pagination_options
from .beanie import apply_to_beanie_query
from .pymongo import get_query_from_options
from .sqlalchemy import apply_to_sqlalchemy_query

__all__ = [
    'apply_to_beanie_query',
    'apply_to_sqlalchemy_query',
    'compute_skip_from_page',
    'get_query_from_options',
    'remove_pagination_options',
]



"""A lightweight, simple and fast parser for OData V4 query options.

Supports standard query parameters and provides helper functions to
apply OData V4 query options to ORM/ODM queries such as SQLAlchemy,
PyMongo and Beanie.

Visit the `repository <https://github.com/daireto/odata-v4-query>`_
for more information.
"""

from .filter_parser import FilterNode, ODataFilterParser
from .filter_tokenizer import ODataFilterTokenizer, Token, TokenType
from .query_parser import ODataQueryOptions, ODataQueryParser

__all__ = [
    'FilterNode',
    'ODataFilterParser',
    'ODataFilterTokenizer',
    'ODataQueryOptions',
    'ODataQueryParser',
    'Token',
    'TokenType',
]

__version__ = '0.4.0'



import pytest
from sqlalchemy import func
from sqlalchemy.orm import Session
from sqlalchemy.sql import select

from odata_v4_query.errors import (
    NoRootClassError,
    UnexpectedNumberOfArgumentsError,
    UnexpectedEmptyArgumentsError,
    UnexpectedNullFiltersError,
    UnexpectedNullFunctionNameError,
    UnexpectedNullOperandError,
    UnexpectedNullOperatorError,
    UnknownFunctionError,
    UnknownOperatorError,
)
from odata_v4_query.filter_parser import FilterNode
from odata_v4_query.query_parser import ODataQueryOptions, ODataQueryParser
from odata_v4_query.utils.sqlalchemy import (
    apply_to_sqlalchemy_query,
    get_query_root_cls,
)

from ._core.sqlalchemy import Post, User, get_engine, seed_data


@pytest.fixture(scope='session')
def session():
    engine = get_engine()
    with Session(engine) as session:
        seed_data(session)
        yield session


class TestSQLAlchemy:
    parser = ODataQueryParser()

    def test_skip(self, session: Session):
        query = select(User)
        users_count = len(session.scalars(query).all())
        options = self.parser.parse_query_string('$skip=2')
        query = apply_to_sqlalchemy_query(options, query)
        result = session.scalars(query).all()
        assert len(result) == users_count - 2
        assert result[0].name == 'Alice'

    def test_top(self, session: Session):
        options = self.parser.parse_query_string('$top=2')
        query = apply_to_sqlalchemy_query(options, User)
        result = session.scalars(query).all()
        assert len(result) == 2
        assert result[0].name == 'John'
        assert result[1].name == 'Jane'

    def test_page(self, session: Session):
        query = select(User)
        users_count = len(session.scalars(query).all())
        options1 = self.parser.parse_query_string('$page=1')
        query1 = apply_to_sqlalchemy_query(options1, User)
        result1 = session.scalars(query1).all()
        options2 = self.parser.parse_query_string('$page=1&$top=4')
        query2 = apply_to_sqlalchemy_query(options2, User)
        result2 = session.scalars(query2).all()
        options3 = self.parser.parse_query_string('$page=2&$top=4')
        query3 = apply_to_sqlalchemy_query(options3, User)
        result3 = session.scalars(query3).all()
        options4 = self.parser.parse_query_string('$page=3&$top=4')
        query4 = apply_to_sqlalchemy_query(options4, User)
        result4 = session.scalars(query4).all()
        options5 = self.parser.parse_query_string('$page=4&$top=4')
        query5 = apply_to_sqlalchemy_query(options5, User)
        result5 = session.scalars(query5).all()
        assert len(result1) == users_count
        assert len(result2) == 4
        assert len(result3) == 4
        assert len(result4) == 2
        assert len(result5) == 0

    def test_filter_comparison_and_logical(self, session: Session):
        options1 = self.parser.parse_query_string(
            "$filter=name eq 'John' and age ge 25"
        )
        query1 = apply_to_sqlalchemy_query(options1, User)
        result1 = session.scalars(query1).all()
        options2 = self.parser.parse_query_string('$filter=age lt 25 or age gt 35')
        query2 = apply_to_sqlalchemy_query(options2, User)
        result2 = session.scalars(query2).all()
        options3 = self.parser.parse_query_string("$filter=name in ('Eve', 'Frank')")
        query3 = apply_to_sqlalchemy_query(options3, User)
        result3 = session.scalars(query3).all()
        options4 = self.parser.parse_query_string("$filter=name nin ('Eve', 'Frank')")
        query4 = apply_to_sqlalchemy_query(options4, User)
        result4 = session.scalars(query4).all()
        options5 = self.parser.parse_query_string(
            "$filter=name ne 'John' and name ne 'Jane'"
        )
        query5 = apply_to_sqlalchemy_query(options5, User)
        result5 = session.scalars(query5).all()
        options6 = self.parser.parse_query_string(
            "$filter=not name eq 'John' and not name eq 'Jane'"
        )
        query6 = apply_to_sqlalchemy_query(options6, User)
        result6 = session.scalars(query6).all()
        assert len(result1) == 1
        assert result1[0].name == 'John'
        assert len(result2) == 4
        assert len(result3) == 2
        assert result3[0].name == 'Eve'
        assert result3[1].name == 'Frank'
        assert len(result4) == 8
        assert len(result5) == 8
        assert len(result6) == 8

    def test_filter_null(self, session: Session):
        options1 = self.parser.parse_query_string('$filter=name eq null')
        query1 = apply_to_sqlalchemy_query(options1, User)
        result1 = session.scalars(query1).all()
        options2 = self.parser.parse_query_string('$filter=name ne null')
        query2 = apply_to_sqlalchemy_query(options2, User)
        result2 = session.scalars(query2).all()
        assert len(result1) == 0
        assert len(result2) == 10

    def test_filter_startswith_function(self, session: Session):
        options = self.parser.parse_query_string(
            "$filter=startswith(name, 'J') and age ge 25"
        )
        query = apply_to_sqlalchemy_query(options, User)
        result = session.scalars(query).all()
        assert len(result) == 2
        assert result[0].name == 'John'
        assert result[1].name == 'Jane'

    def test_filter_endswith_function(self, session: Session):
        options = self.parser.parse_query_string("$filter=endswith(name, 'e')")
        query = apply_to_sqlalchemy_query(options, User)
        result = session.scalars(query).all()
        assert len(result) == 5
        assert result[0].name == 'Jane'
        assert result[1].name == 'Alice'
        assert result[2].name == 'Charlie'
        assert result[3].name == 'Eve'
        assert result[4].name == 'Grace'

    def test_filter_contains_function(self, session: Session):
        options = self.parser.parse_query_string(
            "$filter=contains(name, 'i') and age le 35"
        )
        query = apply_to_sqlalchemy_query(options, User)
        result = session.scalars(query).all()
        assert len(result) == 2
        assert result[0].name == 'Alice'
        assert result[0].age == 35
        assert result[1].name == 'Charlie'
        assert result[1].age == 32

    def test_filter_substring_function(self, session: Session):
        options = self.parser.parse_query_string(
            "$filter=substring(name, 1, 2) eq 'oh'"
        )
        query = apply_to_sqlalchemy_query(options, User)
        result = session.scalars(query).all()
        assert len(result) == 1
        assert result[0].name == 'John'

        options = self.parser.parse_query_string(
            "$filter=substring(name, 1, -1) eq 'lice'"
        )
        query = apply_to_sqlalchemy_query(options, User)
        result = session.scalars(query).all()
        assert len(result) == 1
        assert result[0].name == 'Alice'

    def test_filter_tolower_function(self, session: Session):
        options = self.parser.parse_query_string("$filter=tolower(name) eq 'john'")
        query = apply_to_sqlalchemy_query(options, User)
        result = session.scalars(query).all()
        assert len(result) == 1
        assert result[0].name == 'John'

    def test_filter_toupper_function(self, session: Session):
        options = self.parser.parse_query_string("$filter=toupper(name) eq 'ALICE'")
        query = apply_to_sqlalchemy_query(options, User)
        result = session.scalars(query).all()
        assert len(result) == 1
        assert result[0].name == 'Alice'

    def test_filter_has(self, session: Session):
        options = self.parser.parse_query_string("$filter=addresses has '101 Main St'")
        query = apply_to_sqlalchemy_query(options, User)
        result = session.scalars(query).all()
        assert len(result) == 2
        assert result[0].name == 'Alice'
        assert result[1].name == 'Bob'

    def test_search(self, session: Session):
        options = self.parser.parse_query_string('$search=John')
        query = apply_to_sqlalchemy_query(
            options, User, search_fields=['name', 'email']
        )
        result = session.scalars(query).all()
        assert len(result) == 1
        assert result[0].name == 'John'

    def test_orderby(self, session: Session):
        options = self.parser.parse_query_string('$orderby=name asc,age desc')
        query = apply_to_sqlalchemy_query(options, User)
        result = session.scalars(query).all()
        assert len(result) == 10
        assert result[0].name == 'Alice'
        assert result[1].name == 'Bob'
        assert result[1].age == 40
        assert result[2].name == 'Bob'
        assert result[2].age == 28
        assert result[3].name == 'Charlie'
        assert result[4].name == 'David'
        assert result[5].name == 'Eve'
        assert result[6].name == 'Frank'
        assert result[7].name == 'Grace'
        assert result[8].name == 'Jane'
        assert result[9].name == 'John'

    def test_expand(self, session: Session):
        options = self.parser.parse_query_string('$expand=posts')
        query = apply_to_sqlalchemy_query(options, User)
        result = session.scalars(query).unique().all()
        assert result[0].posts[0].title == 'Post 1'
        assert result[0].posts[1].title == 'Post 2'
        assert result[1].posts[0].title == 'Post 3'
        assert result[1].posts[1].title == 'Post 4'

    def test_count(self, session: Session):
        options = self.parser.parse_query_string('$count=true')
        query = apply_to_sqlalchemy_query(options, User)
        result = session.execute(query).all()
        assert len(result) == 1
        assert result[0][0] == 10

    def test_select(self, session: Session):
        options = self.parser.parse_query_string('$select=name,email')
        query = apply_to_sqlalchemy_query(options, User)
        result = session.execute(query).all()
        assert len(result) == 10
        assert result[0][0] == 'John'
        assert result[0][1] == 'john@example.com'

    def test_filter_nested_field_eq(self, session: Session):
        options = self.parser.parse_query_string("$filter=user/name eq 'John'")
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 2
        assert result[0].title == 'Post 1'
        assert result[1].title == 'Post 2'

    def test_filter_nested_field_comparison(self, session: Session):
        options = self.parser.parse_query_string(
            '$filter=user/age gt 25 and rating ge 3'
        )
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 1
        assert result[0].title == 'Post 3'
        assert result[0].user.name == 'Jane'

    def test_filter_nested_field_startswith(self, session: Session):
        options = self.parser.parse_query_string("$filter=startswith(user/name, 'J')")
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 4
        titles = [r.title for r in result]
        assert 'Post 1' in titles
        assert 'Post 2' in titles
        assert 'Post 3' in titles
        assert 'Post 4' in titles

    def test_filter_nested_field_contains(self, session: Session):
        options = self.parser.parse_query_string("$filter=contains(user/email, 'jane')")
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 2
        assert result[0].title == 'Post 3'
        assert result[1].title == 'Post 4'

    def test_filter_nested_field_endswith(self, session: Session):
        options = self.parser.parse_query_string(
            "$filter=endswith(user/email, 'example.com')"
        )
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 4

    def test_filter_nested_field_tolower(self, session: Session):
        options = self.parser.parse_query_string("$filter=tolower(user/name) eq 'john'")
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 2
        assert result[0].title == 'Post 1'
        assert result[1].title == 'Post 2'

    def test_filter_nested_field_toupper(self, session: Session):
        options = self.parser.parse_query_string("$filter=toupper(user/name) eq 'JANE'")
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 2
        assert result[0].title == 'Post 3'
        assert result[1].title == 'Post 4'

    def test_filter_nested_field_and_operator(self, session: Session):
        options = self.parser.parse_query_string(
            "$filter=user/name eq 'Jane' and rating eq 3"
        )
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 1
        assert result[0].title == 'Post 3'

    def test_filter_multi_level_nested_field_eq(self, session: Session):
        options = self.parser.parse_query_string(
            "$filter=user/profile/address/city eq 'Chicago'"
        )
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 2
        assert result[0].title == 'Post 3'
        assert result[1].title == 'Post 4'

    def test_filter_multi_level_nested_field_comparison(self, session: Session):
        options = self.parser.parse_query_string(
            "$filter=user/profile/address/country eq 'USA'"
        )
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 4

    def test_filter_multi_level_nested_field_startswith(self, session: Session):
        options = self.parser.parse_query_string(
            "$filter=startswith(user/profile/address/city, 'Chi')"
        )
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 2

    def test_filter_multi_level_nested_field_tolower(self, session: Session):
        options = self.parser.parse_query_string(
            "$filter=tolower(user/profile/address/city) eq 'chicago'"
        )
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 2

    def test_filter_multi_level_nested_field_and_operator(self, session: Session):
        options = self.parser.parse_query_string(
            "$filter=user/profile/address/country eq 'USA' and rating ge 4"
        )
        query = apply_to_sqlalchemy_query(options, Post)
        result = session.scalars(query).all()
        assert len(result) == 2
        assert result[0].title == 'Post 1'
        assert result[1].title == 'Post 2'

    def test_unexpected_null_filters(self):
        options = ODataQueryOptions(filter_=FilterNode(type_='value'))
        with pytest.raises(UnexpectedNullFiltersError):
            apply_to_sqlalchemy_query(options, User)

    def test_unexpected_null_operator(self):
        options = ODataQueryOptions(filter_=FilterNode(type_='operator'))
        with pytest.raises(UnexpectedNullOperatorError):
            apply_to_sqlalchemy_query(options, User)

    def test_unexpected_null_operand(self):
        options = ODataQueryOptions(filter_=FilterNode(type_='operator', value='eq'))
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_sqlalchemy_query(options, User)

    def test_unexpected_null_operand_value(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='operator',
                value='eq',
                left=FilterNode(type_='identifier'),
                right=FilterNode(type_='literal', value='John'),
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_sqlalchemy_query(options, User)

    def test_unexpected_null_operand_for_in_nin_operators(self):
        options1 = ODataQueryOptions(
            filter_=FilterNode(
                type_='operator',
                value='in',
                left=FilterNode(type_='identifier', value='name'),
                right=FilterNode(type_='list'),
            )
        )
        options2 = ODataQueryOptions(
            filter_=FilterNode(
                type_='operator',
                value='nin',
                left=FilterNode(type_='identifier', value='name'),
                right=FilterNode(type_='list'),
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_sqlalchemy_query(options1, User)
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_sqlalchemy_query(options2, User)

    def test_unexpected_null_operand_for_has_operator(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='operator',
                value='has',
                left=FilterNode(type_='identifier', value='addresses'),
                right=FilterNode(type_='literal'),
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_sqlalchemy_query(options, User)

    def test_unexpected_null_operand_for_and_or_operators(self):
        options1 = ODataQueryOptions(filter_=FilterNode(type_='operator', value='and'))
        options2 = ODataQueryOptions(filter_=FilterNode(type_='operator', value='or'))
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_sqlalchemy_query(options1, User)
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_sqlalchemy_query(options2, User)

    def test_unexpected_null_operand_for_not_nor_operators(self):
        options1 = ODataQueryOptions(filter_=FilterNode(type_='operator', value='not'))
        options2 = ODataQueryOptions(filter_=FilterNode(type_='operator', value='nor'))
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_sqlalchemy_query(options1, User)
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_sqlalchemy_query(options2, User)

    def test_unknown_operator(self):
        options = ODataQueryOptions(
            filter_=FilterNode(type_='operator', value='unknown')
        )
        with pytest.raises(UnknownOperatorError):
            apply_to_sqlalchemy_query(options, User)

    def test_unexpected_null_function_name(self):
        options = ODataQueryOptions(filter_=FilterNode(type_='function'))
        with pytest.raises(UnexpectedNullFunctionNameError):
            apply_to_sqlalchemy_query(options, User)

    def test_unexpected_empty_arguments(self):
        options = ODataQueryOptions(
            filter_=FilterNode(type_='function', value='startswith')
        )
        with pytest.raises(UnexpectedEmptyArgumentsError):
            apply_to_sqlalchemy_query(options, User)

    def test_unexpected_number_of_arguments(self):
        options1 = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='startswith',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal', value='J'),
                    FilterNode(type_='literal', value='J'),
                ],
            )
        )
        options2 = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='substring',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal', value=1),
                ],
            )
        )
        with pytest.raises(UnexpectedNumberOfArgumentsError):
            apply_to_sqlalchemy_query(options1, User)
        with pytest.raises(UnexpectedNumberOfArgumentsError):
            apply_to_sqlalchemy_query(options2, User)

    def test_unexpected_null_operand_for_function(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='startswith',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal'),
                ],
            )
        )
        with pytest.raises(UnexpectedNullOperandError):
            apply_to_sqlalchemy_query(options, User)

    def test_unknown_function(self):
        options = ODataQueryOptions(
            filter_=FilterNode(
                type_='function',
                value='unknown',
                arguments=[
                    FilterNode(type_='identifier', value='name'),
                    FilterNode(type_='literal', value='J'),
                ],
            )
        )
        with pytest.raises(UnknownFunctionError):
            apply_to_sqlalchemy_query(options, User)

    def test_no_root_class(self):
        query = select(func.count('*'))
        options1 = self.parser.parse_query_string('$filter=name eq null')
        options2 = self.parser.parse_query_string('$search=John')
        options3 = self.parser.parse_query_string('$orderby=name asc')
        options4 = self.parser.parse_query_string('$expand=posts')
        options5 = self.parser.parse_query_string('$select=name,email')
        with pytest.raises(NoRootClassError):
            get_query_root_cls(query, raise_on_none=True)
        with pytest.raises(NoRootClassError):
            apply_to_sqlalchemy_query(options1, query)
        with pytest.raises(NoRootClassError):
            apply_to_sqlalchemy_query(options2, query, search_fields=['name'])
        with pytest.raises(NoRootClassError):
            apply_to_sqlalchemy_query(options3, query)
        with pytest.raises(NoRootClassError):
            apply_to_sqlalchemy_query(options4, query)
        with pytest.raises(NoRootClassError):
            apply_to_sqlalchemy_query(options5, query)



from dataclasses import dataclass
from enum import Enum
from typing import Protocol

from .definitions import FUNCTION_ARITY, OPERATOR_ARITY
from .errors import InvalidNumberError, TokenizeError


class TokenType(str, Enum):
    OPERATOR = 'OPERATOR'
    FUNCTION = 'FUNCTION'
    IDENTIFIER = 'IDENTIFIER'
    LITERAL = 'LITERAL'
    LPAREN = 'LPAREN'
    RPAREN = 'RPAREN'
    COMMA = 'COMMA'


@dataclass
class Token:
    type_: TokenType
    value: str | int | float
    position: int
    args_count: int | None = None


class ODataFilterTokenizerProtocol(Protocol):
    def tokenize(self, expr: str) -> list[Token]: ...


class ODataFilterTokenizer:
    """Tokenizer for OData V4 filter expressions."""

    __position: int
    __tokens: list[Token]

    def tokenize(self, expr: str) -> list[Token]:
        """Convert filter expression string into tokens.

        Parameters
        ----------
        expr : str
            Filter expression to be tokenized.

        Returns
        -------
        list[Token]
            List of tokens extracted from the filter expression.

        Raises
        ------
        TokenizeError
            If the filter expression is invalid.

        Examples
        --------
        >>> from odata_v4_query import ODataFilterTokenizer
        >>> tokenizer = ODataFilterTokenizer()
        >>> tokens = tokenizer.tokenize("name eq 'John' and age gt 25")
        >>> for token in tokens:
        ...     print(token.type, token.value)
        Token(type_=<TokenType.IDENTIFIER: 'IDENTIFIER'>, value='name', position=0)
        Token(type_=<TokenType.OPERATOR: 'OPERATOR'>, value='eq', position=5)
        Token(type_=<TokenType.LITERAL: 'LITERAL'>, value='John', position=8)
        Token(type_=<TokenType.OPERATOR: 'OPERATOR'>, value='and', position=15)
        Token(type_=<TokenType.IDENTIFIER: 'IDENTIFIER'>, value='age', position=19)
        Token(type_=<TokenType.OPERATOR: 'OPERATOR'>, value='gt', position=23)
        Token(type_=<TokenType.LITERAL: 'LITERAL'>, value='25', position=26)

        """
        self.__position = 0
        self.__tokens = []
        expr_len = len(expr)

        char_handlers = {
            '(': lambda: self.__tokens.append(
                Token(TokenType.LPAREN, '(', self.__position),
            ),
            ')': lambda: self.__tokens.append(
                Token(TokenType.RPAREN, ')', self.__position),
            ),
            ',': lambda: self.__tokens.append(
                Token(TokenType.COMMA, ',', self.__position),
            ),
            "'": lambda: self._handle_string_literal(expr),
            '"': lambda: self._handle_string_literal(expr),
        }

        while self.__position < expr_len:
            char = expr[self.__position]

            # skip whitespace
            if char.isspace():
                self.__position += 1
                continue

            # Handle special characters using dispatch dictionary
            handler = char_handlers.get(char)
            if handler:
                handler()
                self.__position += 1
                continue

            # handle numbers (including negative numbers)
            if char.isdigit() or (char == '-' and self.__position + 1 < expr_len and expr[self.__position + 1].isdigit()):
                value, pos = self._extract_number(expr)
                self.__tokens.append(Token(TokenType.LITERAL, value, pos))
                continue

            # handle identifiers, operators, and functions
            if char.isalpha():
                value, pos = self._extract_identifier(expr)
                lowercased_value = value.lower()

                if lowercased_value in OPERATOR_ARITY:
                    self.__tokens.append(
                        Token(
                            TokenType.OPERATOR,
                            lowercased_value,
                            pos,
                            OPERATOR_ARITY[lowercased_value],
                        ),
                    )
                elif lowercased_value in FUNCTION_ARITY:
                    self.__tokens.append(
                        Token(
                            TokenType.FUNCTION,
                            lowercased_value,
                            pos,
                            FUNCTION_ARITY[lowercased_value],
                        ),
                    )
                else:
                    self.__tokens.append(
                        Token(TokenType.IDENTIFIER, value, pos),
                    )
                continue

            raise TokenizeError(char, self.__position)

        return self.__tokens

    def _handle_string_literal(self, expr: str) -> None:
        """Handle string literals.

        Parameters
        ----------
        expr : str
            Expression to extract the string literal from.

        """
        value, pos = self._extract_string_literal(expr)
        self.__tokens.append(Token(TokenType.LITERAL, value, pos))

    def _extract_string_literal(self, expr: str) -> tuple[str, int]:
        """Extract a string literal from the expression.

        Parameters
        ----------
        expr : str
            Expression to extract the string literal from.

        Returns
        -------
        tuple[str, int]
            A tuple containing the string literal and the position of
            the first character of the string literal.

        """
        start_pos = self.__position
        expr_len = len(expr)
        end_pos = start_pos + 1  # skip opening quote
        chars = []

        while end_pos < expr_len:
            char = expr[end_pos]

            # handle escaped characters
            if char == '\\':
                if end_pos + 1 >= expr_len:
                    break

                chars.append(expr[end_pos + 1])
                end_pos += 2
                continue

            # append non-quote characters
            if char not in {"'", '"'}:
                chars.append(char)
                end_pos += 1
                continue

            # end of string literal
            break

        self.__position = end_pos
        return ''.join(chars), start_pos

    def _extract_number(self, expr: str) -> tuple[int | float, int]:
        """Extract a number from the expression.

        Parameters
        ----------
        expr : str
            Expression to extract the number from.

        Returns
        -------
        tuple[int | float, int]
            A tuple containing the number and the position of
            the first character of the number.

        Raises
        ------
        InvalidNumberError
            If the number is invalid.

        """
        start_pos = self.__position
        expr_len = len(expr)
        end_pos = start_pos
        has_decimal = False

        # Handle negative sign
        if end_pos < expr_len and expr[end_pos] == '-':
            end_pos += 1

        while end_pos < expr_len:
            char = expr[end_pos]
            if char.isdigit():
                end_pos += 1
            elif char == '.':
                if has_decimal:
                    raise InvalidNumberError(
                        expr[start_pos:end_pos],
                        start_pos,
                    )

                has_decimal = True
                end_pos += 1
            elif char.isalpha():
                raise InvalidNumberError(expr[start_pos:end_pos], start_pos)
            else:
                break

        self.__position = end_pos

        value = expr[start_pos:end_pos]
        number = float(value) if '.' in value else int(value)
        return number, start_pos

    def _extract_identifier(self, expr: str) -> tuple[str, int]:
        """Extract an identifier from the expression.

        Parameters
        ----------
        expr : str
            Expression to extract the identifier from.

        Returns
        -------
        tuple[str, int]
            A tuple containing the identifier and the position of
            the first character of the identifier.

        """
        start_pos = self.__position
        expr_len = len(expr)
        end_pos = start_pos

        while end_pos < expr_len:
            char = expr[end_pos]
            if not (char.isalnum() or char in '_/'):
                break
            end_pos += 1

        self.__position = end_pos
        return expr[start_pos:end_pos], start_pos



"""SQLAlchemy filter node parser."""

from typing import Any, TypeVar

from sqlalchemy import func, not_
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.sql.elements import BooleanClauseList
from sqlalchemy.sql.operators import (
    OperatorType,
    and_,
    eq,
    ge,
    gt,
    in_op,
    le,
    lt,
    ne,
    notin_op,
    or_,
)

from odata_v4_query.definitions import (
    AND,
    EQ,
    GE,
    GT,
    IN,
    LE,
    LT,
    NE,
    NIN,
    NOR,
    NOT,
    OR,
)
from odata_v4_query.errors import UnknownOperatorError
from odata_v4_query.query_parser import FilterNode

from .base_filter_parser import BaseFilterNodeParser

FilterType = TypeVar('FilterType', bound=DeclarativeBase)

OPERATORS_MAP = {
    EQ: eq,
    NE: ne,
    GT: gt,
    GE: ge,
    LT: lt,
    LE: le,
    IN: in_op,
    NIN: notin_op,
    AND: and_,
    OR: or_,
    NOT: not_,
    NOR: not_,
}


class SQLAlchemyFilterNodeParser(BaseFilterNodeParser):
    """Parser for converting OData filter AST to SQLAlchemy filter.

    See the ``parse()`` method for more information.
    """

    def __init__(self, model: type[FilterType]) -> None:
        """Parser for converting OData filter AST to SQLAlchemy filter.

        Parameters
        ----------
        model : type[FilterType]
            A SQLAlchemy model class.

        """
        super().__init__()
        self.model = model

    def parse(self, filter_node: FilterNode) -> BooleanClauseList:
        """Parse an AST to a SQLAlchemy filter expression.

        Parameters
        ----------
        filter_node : FilterNode
            AST representing the parsed filter expression.

        Returns
        -------
        BooleanClauseList
            SQLAlchemy filter expression.

        Notes
        -----
        The ``has`` and ``nor`` operators are not supported in SQL,
        so they are converted to a LIKE and NOT expressions,
        respectively.

        Examples
        --------
        >>> from odata_v4_query import ODataFilterParser
        >>> from odata_v4_query.utils.filter_parsers import SQLAlchemyFilterNodeParser
        >>> parser = ODataFilterParser()
        >>> ast = parser.parse("name eq 'John' and age gt 25")
        >>> filters = SQLAlchemyFilterNodeParser(User).parse(ast)
        >>> filters
        'users.name = :name_1 AND users.age > :age_1'
        >>> filters.compile(compile_kwargs={'literal_binds': True})
        "users.name = 'John' AND users.age > 25"

        """
        return super().parse(filter_node)

    def node_to_filter_expr(self, filter_node: FilterNode) -> FilterNode:
        if filter_node.type_ == 'function':
            return self.parse_function_node(filter_node)

        if filter_node.type_ == 'operator':
            left = None
            if filter_node.left is not None:
                left = self.node_to_filter_expr(filter_node.left)

                if isinstance(left.value, str):
                    self._current_field_path = left.value
                    left.value = self._resolve_field_path(left.value)

            right = None
            if filter_node.right is not None:
                right = self.node_to_filter_expr(filter_node.right)

            result = self.parse_operator_node(filter_node, left, right)
            self._current_field_path = None
            return result

        return filter_node

    def parse_startswith(self, field: str, value: Any) -> FilterNode:
        column = self._resolve_field_path(field)  # type: ignore
        expr_value = column.ilike(f'{value}%')
        if '/' in field:
            self._current_field_path = field
        result = self._get_value_filter_node(expr_value)
        if '/' in field:
            result.value = self._build_nested_filter(field, result.value)
            self._current_field_path = None
        return result

    def parse_endswith(self, field: str, value: Any) -> FilterNode:
        column = self._resolve_field_path(field)  # type: ignore
        expr_value = column.ilike(f'%{value}')
        if '/' in field:
            self._current_field_path = field
        result = self._get_value_filter_node(expr_value)
        if '/' in field:
            result.value = self._build_nested_filter(field, result.value)
            self._current_field_path = None
        return result

    def parse_contains(self, field: str, value: Any) -> FilterNode:
        column = self._resolve_field_path(field)  # type: ignore
        expr_value = column.ilike(f'%{value}%')
        if '/' in field:
            self._current_field_path = field
        result = self._get_value_filter_node(expr_value)
        if '/' in field:
            result.value = self._build_nested_filter(field, result.value)
            self._current_field_path = None
        return result

    def parse_substring(self, field: str, start: int, length: int) -> FilterNode:
        column = self._resolve_field_path(field)  # type: ignore
        if length < 0:
            expr_value = func.substr(column, start + 1)
        else:
            expr_value = func.substr(column, start + 1, length)
        return self._get_value_filter_node(expr_value)

    def parse_tolower(self, field: str) -> FilterNode:
        column = self._resolve_field_path(field)  # type: ignore
        expr_value = func.lower(column)
        if '/' in field:
            self._current_field_path = field
        return self._get_value_filter_node(expr_value)

    def parse_toupper(self, field: str) -> FilterNode:
        column = self._resolve_field_path(field)  # type: ignore
        expr_value = func.upper(column)
        if '/' in field:
            self._current_field_path = field
        return self._get_value_filter_node(expr_value)

    def parse_membership_operators(
        self,
        left: Any,
        op_node: Any,
        right: Any,
    ) -> FilterNode:
        operator = self._to_sql_operator(op_node)
        return FilterNode(type_='value', value=operator(left, right))

    def parse_comparison_operators(
        self,
        left: Any,
        op_node: Any,
        right: Any,
    ) -> FilterNode:
        if right is None or right == 'null':
            sql_operator = eq(left, None) if op_node == EQ else ne(left, None)
            return FilterNode(type_='value', value=sql_operator)

        operator = self._to_sql_operator(op_node)
        return FilterNode(type_='value', value=operator(left, right))

    def parse_has_operator(self, left: Any, _: Any, right: Any) -> FilterNode:
        return FilterNode(type_='value', value=left.ilike(f'%{right}%'))

    def parse_and_or_operators(self, left: Any, op_node: Any, right: Any) -> FilterNode:
        operator = self._to_sql_operator(op_node)
        value = operator(left, right)
        return FilterNode(type_='value', value=value)

    def parse_not_nor_operators(self, op_node: Any, right: Any) -> FilterNode:
        operator = self._to_sql_operator(op_node)
        value = operator(right)
        return FilterNode(type_='value', value=value)

    def _parse_comparison_or_membership(
        self,
        operator: str,
        left: FilterNode | None,
        right: FilterNode | None,
    ) -> FilterNode:
        """Override to handle nested fields with has()."""
        result = super()._parse_comparison_or_membership(operator, left, right)

        if (
            hasattr(self, '_current_field_path')
            and self._current_field_path
            and '/' in self._current_field_path
        ):
            result.value = self._build_nested_filter(
                self._current_field_path,
                result.value,
            )

        return result

    def _to_sql_operator(self, operator: str) -> OperatorType:
        """Convert an OData operator to a SQLAlchemy operator.

        Parameters
        ----------
        operator : str
            OData operator.

        Returns
        -------
        OperatorType
            SQLAlchemy operator.

        """
        sql_operator = OPERATORS_MAP.get(operator)
        if sql_operator:
            return sql_operator
        raise UnknownOperatorError(operator)  # pragma: no cover

    def _resolve_field_path(self, field_path: str) -> Any:
        """Resolve a field path to a SQLAlchemy column or relationship attribute.

        Converts OData path notation (e.g., 'Customer/Name') to SQLAlchemy
        attribute access (e.g., Customer.Name).

        Parameters
        ----------
        field_path : str
            Field path using OData notation with '/' as separator.

        Returns
        -------
        Any
            SQLAlchemy column or relationship attribute.

        Examples
        --------
        >>> parser._resolve_field_path('name')
        User.name
        >>> parser._resolve_field_path('user/name')
        User.name (from the related User model)

        """
        if '/' not in field_path:
            return getattr(self.model, field_path)

        parts = field_path.split('/')
        current_model = self.model

        for part in parts[:-1]:  # Exclude the field name
            attr = getattr(current_model, part)
            if hasattr(attr.property, 'mapper'):
                current_model = attr.property.mapper.class_
            else:
                current_model = attr  # pragma: no cover

        return getattr(current_model, parts[-1])

    def _build_nested_filter(self, field_path: str, filter_expr: Any) -> Any:
        """Build a nested filter using has() for relationships.

        Supports multiple levels of nesting by wrapping the filter expression
        with has() for each relationship in the path. The method builds the
        nested structure from innermost to outermost by reversing the
        relationship chain.

        Parameters
        ----------
        field_path : str
            Field path using OData notation with '/' as separator
            (e.g., 'user/name', 'user/profile/address/city').
        filter_expr : Any
            The filter expression to apply on the nested field.

        Returns
        -------
        Any
            SQLAlchemy filter expression using has() for relationships.

        Examples
        --------
        Single level: The filter expression is wrapped in has() for the relationship.
        >>> from odata_v4_query.utils.filter_parsers.sql_filter_parser import SQLAlchemyFilterNodeParser
        >>> parser = SQLAlchemyFilterNodeParser(User)
        >>> parser._build_nested_filter('profile/bio', Profile.bio == 'Software Engineer')
        User.has(Profile.bio == 'Software Engineer')

        Multi-level: The relationships are wrapped in reverse order to build the correct
        nested structure from the deepest level outward.
        >>> from odata_v4_query.utils.filter_parsers.sql_filter_parser import SQLAlchemyFilterNodeParser
        >>> parser = SQLAlchemyFilterNodeParser(User)
        >>> parser._build_nested_filter('user/profile/address/city', Address.city == 'NYC')
        User.has(Profile.has(Address.has(Address.city == 'NYC')))

        """
        if '/' not in field_path:
            return filter_expr  # pragma: no cover

        parts = field_path.split('/')
        model_and_relationship_tuples = []
        current_model = self.model

        for part in parts[:-1]:  # Exclude the field name
            relationship = getattr(current_model, part)
            model_and_relationship_tuples.append((current_model, relationship))
            if hasattr(relationship.property, 'mapper'):
                current_model = relationship.property.mapper.class_

        # Build the has() chain from innermost to outermost
        result = filter_expr
        for _, relationship in reversed(model_and_relationship_tuples):
            result = relationship.has(result)

        return result



from typing import Any


class ODataParserError(Exception):
    """Base class for all OData parser errors."""


class AggregationOperatorNotSupportedError(ODataParserError, NotImplementedError):
    """Aggregation operator not supported error."""

    def __init__(self, function_name: str, operator: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        function_name : str
            Function name.
        operator : str
            MongoDB aggregation operator.

        """
        super().__init__(
            f'Found unsupported aggregation operator {operator} in {function_name!r}.',
        )


class InvalidOrderDirectionError(ODataParserError, ValueError):
    """Invalid order direction error."""

    def __init__(self, direction: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        direction : str
            Direction.

        """
        super().__init__(f'invalid order direction {direction!r}')


class InvalidNumberError(ODataParserError, ValueError):
    """Invalid number error."""

    def __init__(self, value: str, position: int) -> None:
        """Initialize the error.

        Parameters
        ----------
        value : str
            Value.
        position : int
            Start position.

        """
        super().__init__(
            f'invalid number at position {position}, got {value!r}',
        )


class OpeningParenthesisExpectedError(ODataParserError, ValueError):
    """Opening parenthesis expected error."""

    def __init__(self, func_name: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        func_name : str
            Function name.

        """
        super().__init__(f"expected '(' after {func_name!r} function name")


class CommaOrClosingParenthesisExpectedError(ODataParserError, ValueError):
    """Comma or closing parenthesis expected error."""

    def __init__(self, value: str, expected_after: bool = False) -> None:
        """Initialize the error.

        Parameters
        ----------
        value : str
            Value.
        expected_after : bool, optional
            Whether the comma or closing parenthesis is expected
            after the value, by default False.

        """
        value = f' after {value!r}' if expected_after else f', got {value!r}'
        super().__init__(f"expected ',' or ')'{value}")


class MissingClosingParenthesisError(ODataParserError, ValueError):
    """Missing closing parenthesis error."""

    def __init__(self) -> None:
        """Initialize the error."""
        super().__init__('missing closing parenthesis')


class NoNumericValueError(ODataParserError, ValueError):
    """No numeric value error."""

    def __init__(self, value: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        value : str
            Value.

        """
        super().__init__(f'expected a numeric value, got {value!r}')


class NoPositiveError(ODataParserError, ValueError):
    """No positive integer value error."""

    def __init__(self, param: str, value: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        param : str
            Parameter.
        value : str
            Value.

        """
        super().__init__(f'expected {param} to be a positive integer, got {value!r}')


class NoRootClassError(ODataParserError, ValueError):
    """No root class found error."""

    def __init__(self, query: str, option_name: str | None = None) -> None:
        """Initialize the error.

        Parameters
        ----------
        query : str
            Query.
        option_name : str | None, optional
            Option name, by default None.

        """
        message = f'could not find root class of query {query!r}'
        if option_name:
            message += f' for option {option_name!r}'
        super().__init__(message)


class TokenizeError(ODataParserError, ValueError):
    """Tokenizer error."""

    def __init__(self, char: str, position: int) -> None:
        """Initialize the error.

        Parameters
        ----------
        char : str
            Character.
        position : int
            Character position.

        """
        super().__init__(f'unexpected character {char!r} at position {position}')


class UnexpectedEmptyArgumentsError(ODataParserError, ValueError):
    """Unexpected empty arguments error."""

    def __init__(self, function_name: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        function_name : str
            Function name.

        """
        super().__init__(f'unexpected empty arguments for function {function_name!r}')


class UnexpectedEndOfExpressionError(ODataParserError, ValueError):
    """Unexpected end of expression error."""

    def __init__(self) -> None:
        """Initialize the error."""
        super().__init__('unexpected end of expression')


class UnexpectedNullFiltersError(ODataParserError, ValueError):
    """Unexpected null filters error."""

    def __init__(self, node_repr: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        node_repr : str
            Node representation.

        """
        super().__init__(f'unexpected null filters in {node_repr!r}')


class UnexpectedNullFunctionNameError(ODataParserError, ValueError):
    """Unexpected null function name error."""

    def __init__(self, node_repr: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        node_repr : str
            Node representation.

        """
        super().__init__(f'unexpected null function name in {node_repr!r}')


class UnexpectedNullIdentifierError(ODataParserError, ValueError):
    """Unexpected null identifier error."""

    def __init__(self, node_repr: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        node_repr : str
            Node representation.

        """
        super().__init__(f'unexpected null identifier in {node_repr!r}')


class UnexpectedNullListError(ODataParserError, ValueError):
    """Unexpected null list error."""

    def __init__(self, node_repr: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        node_repr : str
            Node representation.

        """
        super().__init__(f'unexpected null list in {node_repr!r}')


class UnexpectedNullLiteralError(ODataParserError, ValueError):
    """Unexpected null literal error."""

    def __init__(self, node_repr: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        node_repr : str
            Node representation.

        """
        super().__init__(f'unexpected null literal in {node_repr!r}')


class UnexpectedNullNodeTypeError(ODataParserError, ValueError):
    """Unexpected null node type error."""

    def __init__(self, node_repr: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        node_repr : str
            Node representation.

        """
        super().__init__(f'unexpected null node type in {node_repr!r}')


class UnexpectedNullOperandError(ODataParserError, ValueError):
    """Unexpected null operand error."""

    def __init__(self, operator_or_function: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        operator_or_function : str
            Operator or function.

        """
        super().__init__(
            f'unexpected null operand for {operator_or_function!r}',
        )


class UnexpectedNullOperatorError(ODataParserError, ValueError):
    """Unexpected null operator error."""

    def __init__(self, node_repr: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        node_repr : str
            Node representation.

        """
        super().__init__(f'unexpected null operator in {node_repr!r}')


class UnexpectedNumberOfArgumentsError(ODataParserError, ValueError):
    """Unexpected number of arguments error."""

    def __init__(self, function_name: str, expected: int, got: int) -> None:
        """Initialize the error.

        Parameters
        ----------
        function_name : str
            Function name.
        expected : int
            Expected number of arguments.
        got : int
            Got number of arguments.

        """
        super().__init__(
            f'expected {expected} arguments for function {function_name!r}, got {got}',
        )


class UnexpectedTokenError(ODataParserError, ValueError):
    """Unexpected token error."""

    def __init__(self, token: str, position: int) -> None:
        """Initialize the error.

        Parameters
        ----------
        token : str
            Token.
        position : int
            Token position.

        """
        super().__init__(f'unexpected token {token!r} at position {position}')


class UnexpectedTypeError(ODataParserError, ValueError):
    """Unexpected type error."""

    def __init__(self, function_name: str, expected_type: str, got: Any) -> None:
        """Initialize the error.

        Parameters
        ----------
        function_name : str
            Function name.
        expected_type : str
            Expected type.
        got : Any
            Got value.

        """
        super().__init__(
            f'expected {expected_type!r} for function {function_name!r}, got {got!r}',
        )


class UnknownFunctionError(ODataParserError, ValueError):
    """Unknown function error."""

    def __init__(self, function_name: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        function_name : str
            Function name.

        """
        super().__init__(f'unknown function {function_name!r}')


class UnknownNodeTypeError(ODataParserError, ValueError):
    """Unknown node type error."""

    def __init__(self, node_type: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        node_type : str
            Node type.

        """
        super().__init__(f'unknown node type {node_type!r}')


class UnknownOperatorError(ODataParserError, ValueError):
    """Unknown operator error."""

    def __init__(self, operator: str) -> None:
        """Initialize the error.

        Parameters
        ----------
        operator : str
            Operator.

        """
        super().__init__(f'unknown operator {operator!r}')


class UnsupportedFormatError(ODataParserError, ValueError):
    """Initialize the error."""

    def __init__(self, fmt: str) -> None:
        """Unsupported format error.

        Parameters
        ----------
        fmt : str
            Format.

        """
        super().__init__(f'unsupported format {fmt!r}')



import pytest

from odata_v4_query.errors import (
    CommaOrClosingParenthesisExpectedError,
    MissingClosingParenthesisError,
    OpeningParenthesisExpectedError,
    UnexpectedEndOfExpressionError,
    UnexpectedNullFunctionNameError,
    UnexpectedNullIdentifierError,
    UnexpectedNullListError,
    UnexpectedNullLiteralError,
    UnexpectedNullNodeTypeError,
    UnexpectedNullOperandError,
    UnexpectedNullOperatorError,
    UnexpectedTokenError,
    UnknownNodeTypeError,
)
from odata_v4_query.filter_parser import FilterNode, ODataFilterParser


class TestFilterParser:
    parser = ODataFilterParser()

    def test_parse_ast(self):
        ast = self.parser.parse("name eq 'John' and age gt 25")
        assert ast.type_ == 'operator'
        assert ast.value == 'and'
        assert ast.left is not None
        assert ast.left.type_ == 'operator'
        assert ast.left.value == 'eq'
        assert ast.left.left is not None
        assert ast.left.left.type_ == 'identifier'
        assert ast.left.left.value == 'name'
        assert ast.left.right is not None
        assert ast.left.right.type_ == 'literal'
        assert ast.left.right.value == 'John'
        assert ast.right is not None
        assert ast.right.type_ == 'operator'
        assert ast.right.value == 'gt'
        assert ast.right.left is not None
        assert ast.right.left.type_ == 'identifier'
        assert ast.right.left.value == 'age'
        assert ast.right.right is not None
        assert ast.right.right.type_ == 'literal'
        assert ast.right.right.value == 25

    def test_parse_empty_ast(self):
        ast = self.parser.parse('')
        assert ast.type_ == 'value'
        assert ast.value is None

    def test_parse_in(self):
        ast = self.parser.parse("name in ('John', 'Jane')")
        assert ast.type_ == 'operator'
        assert ast.value == 'in'
        assert ast.left is not None
        assert ast.left.type_ == 'identifier'
        assert ast.left.value == 'name'
        assert ast.right is not None
        assert ast.right.type_ == 'list'
        assert ast.right.arguments is not None
        assert ast.right.arguments[0].type_ == 'literal'
        assert ast.right.arguments[0].value == 'John'
        assert ast.right.arguments[1].type_ == 'literal'
        assert ast.right.arguments[1].value == 'Jane'

    def test_parse_in_empty_list(self):
        ast = self.parser.parse('name in ()')
        assert ast.type_ == 'operator'
        assert ast.value == 'in'
        assert ast.left is not None
        assert ast.left.type_ == 'identifier'
        assert ast.left.value == 'name'
        assert ast.right is not None
        assert ast.right.type_ == 'list'
        assert ast.right.arguments == []

    def test_parse_null_identifier(self):
        ast = self.parser.parse('null')
        assert ast.type_ == 'identifier'
        assert ast.value is None

    def test_parse_null_identifier_disabled(self):
        parser = ODataFilterParser(parse_null_identifier=False)
        ast = parser.parse('null')
        assert ast.type_ == 'identifier'
        assert ast.value == 'null'

    def test_parse_function(self):
        ast = self.parser.parse("startswith(name, 'J')")
        assert ast.type_ == 'function'
        assert ast.value == 'startswith'
        assert ast.arguments is not None
        assert ast.arguments[0].type_ == 'identifier'
        assert ast.arguments[0].value == 'name'
        assert ast.arguments[1].type_ == 'literal'
        assert ast.arguments[1].value == 'J'

    def test_parse_not(self):
        ast = self.parser.parse("not name eq 'John'")
        assert ast.type_ == 'operator'
        assert ast.value == 'not'
        assert ast.right is not None
        assert ast.right.type_ == 'operator'
        assert ast.right.value == 'eq'
        assert ast.right.left is not None
        assert ast.right.left.type_ == 'identifier'
        assert ast.right.left.value == 'name'
        assert ast.right.right is not None
        assert ast.right.right.type_ == 'literal'
        assert ast.right.right.value == 'John'

    def test_no_primary_expression(self):
        with pytest.raises(UnexpectedTokenError):
            self.parser.parse('eq')

    def test_no_opening_parenthesis(self):
        with pytest.raises(OpeningParenthesisExpectedError):
            self.parser.parse('startswith')

    def test_missing_comma_or_closing_parenthesis(self):
        with pytest.raises(CommaOrClosingParenthesisExpectedError):
            self.parser.parse("name in ('John', 'Jane'")

        with pytest.raises(CommaOrClosingParenthesisExpectedError):
            self.parser.parse("name in ('John'd")

    def test_unexpected_end_of_expression(self):
        with pytest.raises(UnexpectedEndOfExpressionError):
            self.parser.parse('name in')

    def test_missing_closing_parenthesis(self):
        with pytest.raises(MissingClosingParenthesisError):
            self.parser.parse('name in (')

    def test_evaluate(self):
        ast1 = self.parser.parse("name eq 'John' and age gt 25")
        ast2 = self.parser.parse('name eq null')
        ast3 = self.parser.parse("startswith(name, 'John')")
        ast4 = self.parser.parse("name in ('John', 'Jane')")
        ast5 = self.parser.parse("not name eq 'John'")
        assert self.parser.evaluate(ast1) == "name eq 'John' and age gt 25"
        assert self.parser.evaluate(ast2) == 'name eq null'
        assert self.parser.evaluate(ast3) == "startswith(name, 'John')"
        assert self.parser.evaluate(ast4) == "name in ('John', 'Jane')"
        assert self.parser.evaluate(ast5) == "not name eq 'John'"

    def test_unexpected_null_node_type(self):
        with pytest.raises(UnexpectedNullNodeTypeError):
            self.parser.evaluate(FilterNode(type_=None))  # type: ignore

    def test_unknown_node_type(self):
        with pytest.raises(UnknownNodeTypeError):
            self.parser.evaluate(FilterNode(type_='unknown'))  # type: ignore

    def test_unexpected_null_literal(self):
        with pytest.raises(UnexpectedNullLiteralError):
            self.parser.evaluate(FilterNode(type_='literal'))

    def test_unexpected_null_identifier(self):
        parser = ODataFilterParser(parse_null_identifier=False)
        assert self.parser.evaluate(FilterNode(type_='identifier')) == 'null'
        with pytest.raises(UnexpectedNullIdentifierError):
            parser.evaluate(FilterNode(type_='identifier'))

    def test_unexpected_null_list(self):
        with pytest.raises(UnexpectedNullListError):
            self.parser.evaluate(FilterNode(type_='list'))

    def test_unexpected_null_operator(self):
        with pytest.raises(UnexpectedNullOperatorError):
            self.parser.evaluate(FilterNode(type_='operator'))

    def test_unexpected_null_operand_for_not_operator(self):
        with pytest.raises(UnexpectedNullOperandError):
            self.parser.evaluate(FilterNode(type_='operator', value='not'))

    def test_unexpected_null_operand_for_operator(self):
        with pytest.raises(UnexpectedNullOperandError):
            self.parser.evaluate(FilterNode(type_='operator', value='eq'))

    def test_unexpected_null_function_name(self):
        with pytest.raises(UnexpectedNullFunctionNameError):
            self.parser.evaluate(FilterNode(type_='function'))

    def test_unexpected_empty_arguments(self):
        with pytest.raises(UnexpectedNullListError):
            self.parser.evaluate(FilterNode(type_='function', value='startswith'))



"""MongoDB filter node parser."""

from typing import Any

from odata_v4_query.definitions import (
    AND,
    EQ,
    GE,
    GT,
    IN,
    LE,
    LT,
    NE,
    NIN,
    NOR,
    NOT,
    OR,
)
from odata_v4_query.errors import (
    AggregationOperatorNotSupportedError,
    UnknownOperatorError,
)
from odata_v4_query.query_parser import FilterNode

from .base_filter_parser import BaseFilterNodeParser


class MongoDBFilterNodeParser(BaseFilterNodeParser):
    """Parser for converting OData filter AST to MongoDB filter.

    See the ``parse()`` method for more information.
    """

    def parse(self, filter_node: FilterNode) -> dict[str, Any]:
        """Parse an AST to a MongoDB filter.

        Parameters
        ----------
        filter_node : FilterNode
            AST representing the parsed filter expression.

        Returns
        -------
        dict[str, Any]
            MongoDB filter expression.

        Examples
        --------
        >>> from odata_v4_query import ODataFilterParser
        >>> from odata_v4_query.utils.filter_parsers import MongoDBFilterNodeParser
        >>> parser = ODataFilterParser()
        >>> ast = parser.parse("name eq 'John' and age gt 25")
        >>> MongoDBFilterNodeParser().parse(ast)
        {'$and': [{'name': {'$eq': 'John'}}, {'age': {'$gt': 25}}]}

        """
        return super().parse(filter_node)

    def parse_startswith(self, field: str, value: Any) -> FilterNode:
        field = self._normalize_field_path(field)
        expr_value = {
            field: {
                '$regex': f'^{value}',
                '$options': 'i',
            },
        }
        return self._get_value_filter_node(expr_value)

    def parse_endswith(self, field: str, value: Any) -> FilterNode:
        field = self._normalize_field_path(field)
        expr_value = {
            field: {
                '$regex': f'{value}$',
                '$options': 'i',
            },
        }
        return self._get_value_filter_node(expr_value)

    def parse_contains(self, field: str, value: Any) -> FilterNode:
        field = self._normalize_field_path(field)
        expr_value = {
            field: {
                '$regex': value,
                '$options': 'i',
            },
        }
        return self._get_value_filter_node(expr_value)

    def parse_substring(self, *_) -> FilterNode:
        func_name = 'substring'
        op = '$substr'
        raise AggregationOperatorNotSupportedError(func_name, op)

    def parse_tolower(self, *_) -> FilterNode:
        func_name = 'tolower'
        op = '$toLower'
        raise AggregationOperatorNotSupportedError(func_name, op)

    def parse_toupper(self, *_) -> FilterNode:
        func_name = 'toupper'
        op = '$toUpper'
        raise AggregationOperatorNotSupportedError(func_name, op)

    def parse_membership_operators(
        self,
        left: Any,
        op_node: Any,
        right: Any,
    ) -> FilterNode:
        operator = self._to_mongo_operator(op_node)
        field = self._normalize_field_path(left) if isinstance(left, str) else left
        return FilterNode(type_='value', value={field: {operator: right}})

    def parse_comparison_operators(
        self,
        left: Any,
        op_node: Any,
        right: Any,
    ) -> FilterNode:
        operator = self._to_mongo_operator(op_node)
        field = self._normalize_field_path(left) if isinstance(left, str) else left
        return FilterNode(type_='value', value={field: {operator: right}})

    def parse_has_operator(self, left: Any, _: Any, right: Any) -> FilterNode:
        field = self._normalize_field_path(left) if isinstance(left, str) else left
        return FilterNode(type_='value', value={field: right})

    def parse_and_or_operators(self, left: Any, op_node: Any, right: Any) -> FilterNode:
        operator = self._to_mongo_operator(op_node)
        value = {operator: [left, right]}
        return FilterNode(type_='value', value=value)

    def parse_not_nor_operators(self, op_node: Any, right: Any) -> FilterNode:
        operator = self._to_mongo_operator(op_node)
        field, comparison = right.popitem()
        value = {field: {operator: comparison}}
        return FilterNode(type_='value', value=value)

    def _to_mongo_operator(self, operator: str) -> str:
        """Convert an OData operator to a MongoDB operator.

        Parameters
        ----------
        operator : str
            OData operator.

        Returns
        -------
        str
            MongoDB operator.

        """
        if operator == GE:
            return '$gte'
        if operator == LE:
            return '$lte'
        if operator in (EQ, NE, GT, LT, IN, NIN, AND, OR, NOT, NOR):
            return f'${operator}'
        raise UnknownOperatorError(operator)  # pragma: no cover

    def _normalize_field_path(self, field: str) -> str:
        """Convert OData path separator (/) to MongoDB dot notation (.).

        Parameters
        ----------
        field : str
            Field path using OData notation (e.g., 'Customer/Name').

        Returns
        -------
        str
            Field path using MongoDB dot notation (e.g., 'Customer.Name').

        """
        return field.replace('/', '.')



import pytest

from odata_v4_query.errors import InvalidNumberError, TokenizeError
from odata_v4_query.filter_tokenizer import ODataFilterTokenizer


class TestFilterTokenizer:
    tokenizer = ODataFilterTokenizer()

    def test_tokenize(self):
        tokens1 = self.tokenizer.tokenize("name eq 'John' and age gt 25")
        tokens2 = self.tokenizer.tokenize('name eq "John" and age gt 25')
        tokens3 = self.tokenizer.tokenize("name eq 'D\\'Angelo' and age gt 25")
        tokens4 = self.tokenizer.tokenize("name eq 'D\\")
        tokens5 = self.tokenizer.tokenize(
            "name eq 'John' or startswith(name, 'J') and age gt 25 or age in (25, 30)"
        )
        assert len(tokens1) == 7
        assert len(tokens2) == 7
        assert len(tokens3) == 7
        assert len(tokens4) == 3
        assert len(tokens5) == 22

    def test_invalid_character(self):
        with pytest.raises(TokenizeError):
            self.tokenizer.tokenize("name eq 'John' and age gt #")

    def test_invalid_number(self):
        with pytest.raises(InvalidNumberError):
            self.tokenizer.tokenize("name eq 'John' and age gt 25d")

    def test_invalid_number_with_more_than_one_decimal_point(self):
        with pytest.raises(InvalidNumberError):
            self.tokenizer.tokenize("name eq 'John' and age gt 24..0")



import copy
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from typing import Literal
from urllib.parse import parse_qs, urlparse

from .definitions import DEFAULT_FORMAT_OPTIONS, DEFAULT_ORDERBY_DIRECTION
from .errors import (
    InvalidOrderDirectionError,
    NoNumericValueError,
    NoPositiveError,
    UnsupportedFormatError,
)
from .filter_parser import (
    FilterNode,
    ODataFilterParser,
    ODataFilterParserProtocol,
)

OptionCallback = Callable[[str, 'ODataQueryOptions'], None]


@dataclass
class OrderbyItem:
    field: str
    direction: Literal['asc', 'desc']


@dataclass
class ODataQueryOptions:
    count: bool = False
    expand: list[str] | None = None
    filter_: FilterNode | None = None
    format_: str | None = None
    orderby: list[OrderbyItem] | None = None
    search: str | None = None
    select: list[str] | None = None
    skip: int | None = None
    top: int | None = None
    page: int | None = None

    def clone(self, deep: bool = False) -> 'ODataQueryOptions':
        """Clone the query options.

        Parameters
        ----------
        deep : bool, optional
            Whether to perform a deep copy using ``copy.deepcopy()``,
            by default nested objects (expand, filter_, orderby, select)
            are copied shallowly. **Use with caution:** Deep copying large
            lists and ASTs can be very expensive, and lead to high overhead.

        Returns
        -------
        ODataQueryOptions
            Cloned query options.

        """
        return ODataQueryOptions(
            count=self.count,
            expand=copy.deepcopy(self.expand) if deep else self.expand,
            filter_=copy.deepcopy(self.filter_) if deep else self.filter_,
            format_=self.format_,
            orderby=copy.deepcopy(self.orderby) if deep else self.orderby,
            search=self.search,
            select=copy.deepcopy(self.select) if deep else self.select,
            skip=self.skip,
            top=self.top,
            page=self.page,
        )


class ODataQueryParser:
    """Parser for standard OData V4 query options."""

    __supported_options: dict[str, OptionCallback]
    __supported_formats: Sequence[str]
    __filter_parser: ODataFilterParserProtocol

    def __init__(
        self,
        supported_options: dict[str, OptionCallback] | None = None,
        supported_formats: Sequence[str] | None = None,
        filter_parser: ODataFilterParserProtocol | None = None,
    ) -> None:
        """Initialize the query parser.

        Parameters
        ----------
        supported_options : dict[str, OptionCallback] | None, optional
            Dictionary of supported query options, by default None.
        supported_formats : Sequence[str] | None, optional
            Sequence of supported response formats, by default None.
        filter_parser : ODataFilterParserProtocol | None, optional
            Filter parser, by default, it uses an instance
            of ``ODataFilterParser``.

        """
        self.__supported_options = supported_options or {
            '$count': self._parse_count,
            '$expand': self._parse_expand,
            '$filter': self._parse_filter,
            '$format': self._parse_format,
            '$orderby': self._parse_orderby,
            '$search': self._parse_search,
            '$select': self._parse_select,
            '$skip': self._parse_skip,
            '$top': self._parse_top,
            '$page': self._parse_page,
        }
        self.__supported_formats = supported_formats or DEFAULT_FORMAT_OPTIONS
        self.__filter_parser = filter_parser or ODataFilterParser()

    def set_supported_options(
        self,
        supported_options: dict[str, OptionCallback],
    ) -> None:
        """Set the supported query options.

        Parameters
        ----------
        supported_options : dict[str, OptionCallback]
            Dictionary of supported query options.

        """
        self.__supported_options = supported_options  # pragma: no cover

    def set_supported_formats(self, supported_formats: Sequence[str]) -> None:
        """Set the supported response formats.

        Parameters
        ----------
        supported_formats : Sequence[str]
            Sequence of supported response formats.

        """
        self.__supported_formats = supported_formats  # pragma: no cover

    def set_filter_parser(
        self,
        filter_parser: ODataFilterParserProtocol,
    ) -> None:
        """Set the filter parser.

        Parameters
        ----------
        filter_parser : ODataFilterParserProtocol
            Filter parser.

        """
        self.__filter_parser = filter_parser  # pragma: no cover

    def parse_url(self, url: str) -> ODataQueryOptions:
        """Parse a complete OData URL and extracts query options.

        Parameters
        ----------
        url : str
            Complete OData URL including query parameters.

        Returns
        -------
        ODataQueryOptions
            Parsed parameters.

        Examples
        --------
        >>> from odata_v4_query import ODataQueryParser
        >>> parser = ODataQueryParser()
        >>> url = 'https://example.com/odata?$count=true&$top=10&$skip=20'
        >>> options = parser.parse_url(url)
        >>> options.count
        True
        >>> options.top
        10
        >>> options.skip
        20

        """
        parsed_url = urlparse(url)
        return self.parse_query_string(parsed_url.query)

    def parse_query_string(self, query_string: str) -> ODataQueryOptions:
        """Parse a complete OData query string and extracts query options.

        Parameters
        ----------
        query_string : str
            Complete OData query string including query parameters.

        Returns
        -------
        ODataQueryOptions
            Parsed parameters.

        Examples
        --------
        >>> from odata_v4_query import ODataQueryParser
        >>> parser = ODataQueryParser()
        >>> query_string = '$count=true&$top=10&$skip=20'
        >>> options = parser.parse_query_string(query_string)
        >>> options.count
        True
        >>> options.top
        10
        >>> options.skip
        20

        """
        query_params = parse_qs(query_string)
        return self.parse_query_params(query_params)

    def parse_query_params(
        self,
        query_params: dict[str, list[str]],
    ) -> ODataQueryOptions:
        """Parse OData query parameters and returns structured options.

        Parameters
        ----------
        query_params : dict[str, list[str]]
            Dictionary of query parameters.

        Returns
        -------
        ODataQueryOptions
            Parsed parameters.

        Examples
        --------
        >>> from odata_v4_query import ODataQueryParser
        >>> parser = ODataQueryParser()
        >>> query_params = {'$count': ['true'], '$top': ['10'], '$skip': ['20']}
        >>> options = parser.parse_query_params(query_params)
        >>> options.count
        True
        >>> options.top
        10
        >>> options.skip
        20

        """
        options = ODataQueryOptions()

        for param, values in query_params.items():
            if param in self.__supported_options:
                # get the first value since OData parameters shouldn't
                # have multiple values
                value = values[0] if values else None
                if not value:
                    continue

                parser_func = self.__supported_options[param]
                parser_func(value, options)

        return options

    def evaluate(
        self,
        options_or_filter: ODataQueryOptions | FilterNode,
    ) -> str:
        """Evaluate an AST to the original filter expression.

        Parameters
        ----------
        options_or_filter : ODataQueryOptions | FilterNode
            Parsed query options or filter AST.

        Returns
        -------
        str
            Filter expression.

        Examples
        --------
        >>> from odata_v4_query import ODataQueryParser
        >>> parser = ODataQueryParser()
        >>> query_string = "$filter=name eq 'John' and age gt 25"
        >>> ast = parser.parse_query_string(query_string)
        >>> parser.evaluate(ast)
        "name eq 'John' and age gt 25"

        """
        if isinstance(options_or_filter, FilterNode):
            return self.__filter_parser.evaluate(options_or_filter)

        if options_or_filter.filter_:
            return self.__filter_parser.evaluate(options_or_filter.filter_)

        return ''

    def _parse_count(self, value: str, options: ODataQueryOptions) -> None:
        """Parse $count parameter.

        Parameters
        ----------
        value : str
            Value of the parameter.
        options : ODataQueryOptions
            Current query options object.

        """
        options.count = value.lower() == 'true'

    def _parse_expand(self, value: str, options: ODataQueryOptions) -> None:
        """Parse $expand parameter.

        Parameters
        ----------
        value : str
            Value of the parameter.
        options : ODataQueryOptions
            Current query options object.

        """
        options.expand = self._split_by_comma(value)

    def _parse_filter(self, value: str, options: ODataQueryOptions) -> None:
        """Parse $filter parameter.

        Parameters
        ----------
        value : str
            Value of the parameter.
        options : ODataQueryOptions
            Current query options object.

        """
        options.filter_ = self.__filter_parser.parse(value)

    def _parse_format(self, value: str, options: ODataQueryOptions) -> None:
        """Parse $format parameter.

        Parameters
        ----------
        value : str
            Value of the parameter.
        options : ODataQueryOptions
            Current query options object.

        Raises
        ------
        UnsupportedFormatError
            If the format is not supported.

        """
        if value not in self.__supported_formats:
            raise UnsupportedFormatError(value)

        options.format_ = value

    def _parse_orderby(self, value: str, options: ODataQueryOptions) -> None:
        """Parse $orderby parameter.

        Parameters
        ----------
        value : str
            Value of the parameter.
        options : ODataQueryOptions
            Current query options object.

        """
        orderby_list = []
        for item in value.split(','):
            item_and_direction = item.strip()
            if not item_and_direction:
                continue

            if ' ' in item_and_direction:
                field, direction = item_and_direction.rsplit(maxsplit=1)
            else:
                field, direction = item_and_direction, DEFAULT_ORDERBY_DIRECTION

            direction = direction.lower()
            if direction not in ('asc', 'desc'):
                raise InvalidOrderDirectionError(direction)

            orderby_list.append(
                OrderbyItem(
                    field=field.strip(),
                    direction=direction.strip(),  # type: ignore
                ),
            )

        options.orderby = orderby_list

    def _parse_search(self, value: str, options: ODataQueryOptions) -> None:
        """Parse $search parameter.

        Parameters
        ----------
        value : str
            Value of the parameter.
        options : ODataQueryOptions
            Current query options object.

        """
        options.search = value.strip()

    def _parse_select(self, value: str, options: ODataQueryOptions) -> None:
        """Parse $select parameter.

        Parameters
        ----------
        value : str
            Value of the parameter.
        options : ODataQueryOptions
            Current query options object.

        """
        options.select = self._split_by_comma(value)

    def _parse_skip(self, value: str, options: ODataQueryOptions) -> None:
        """Parse $skip parameter.

        Parameters
        ----------
        value : str
            Value of the parameter.
        options : ODataQueryOptions
            Current query options object.

        Raises
        ------
        NoNumericValueError
            If the value is not a number.
        NoPositiveError
            If the value is not a positive integer.

        """
        try:
            options.skip = int(value)
        except (ValueError, TypeError) as e:
            raise NoNumericValueError(value) from e

        if options.skip < 0:
            raise NoPositiveError(param='$skip', value=value)

    def _parse_top(self, value: str, options: ODataQueryOptions) -> None:
        """Parse $top parameter.

        Parameters
        ----------
        value : str
            Value of the parameter.
        options : ODataQueryOptions
            Current query options object.

        Raises
        ------
        NoNumericValueError
            If the value is not a number.
        NoPositiveError
            If the value is not a positive integer.

        """
        try:
            options.top = int(value)
        except (ValueError, TypeError) as e:
            raise NoNumericValueError(value) from e

        if options.top < 0:
            raise NoPositiveError(param='$top', value=value)

    def _parse_page(self, value: str, options: ODataQueryOptions) -> None:
        """Parse $page parameter.

        Parameters
        ----------
        value : str
            Value of the parameter.
        options : ODataQueryOptions
            Current query options object.

        Raises
        ------
        NoNumericValueError
            If the value is not a number.
        NoPositiveError
            If the value is not a positive integer.

        """
        try:
            options.page = int(value)
        except (ValueError, TypeError) as e:
            raise NoNumericValueError(value) from e

        if options.page < 0:
            raise NoPositiveError(param='$page', value=value)

    def _split_by_comma(self, value: str) -> list[str]:
        """Split a string by comma.

        Parameters
        ----------
        value : str
            String to be split.

        Returns
        -------
        list[str]
            List of strings.

        """
        return [item.strip() for item in value.split(',')]



from odata_v4_query.definitions import DEFAULT_LIMIT
from odata_v4_query.query_parser import ODataQueryOptions


def compute_skip_from_page(
    options: ODataQueryOptions,
    default_limit: int = DEFAULT_LIMIT,
) -> None:
    """Compute $skip from $page if $page is provided.

    Changes are applied in-place.

    If ``$top`` is not provided, it defaults to
    ``odata_v4_query.definitions.DEFAULT_LIMIT``.

    The ``$skip`` is computed as ``(page - 1) * top``.
    If it is provided, it is overwritten.

    Parameters
    ----------
    options : ODataQueryOptions
        Parsed query options.
    default_limit : int, optional
        Default limit to use if ``$top`` is not provided,
        by default ``odata_v4_query.definitions.DEFAULT_LIMIT``.

    """
    if options.page:
        options.top = options.top or default_limit
        options.skip = (options.page - 1) * options.top


def remove_pagination_options(options: ODataQueryOptions) -> None:
    """Remove pagination options from query.

    Changes are applied in-place.

    Parameters
    ----------
    options : ODataQueryOptions
        Parsed query options.

    """
    options.top = None
    options.skip = None
    options.page = None



"""Utility function for applying OData query options to a PyMongo query.

See ``get_query_from_options()`` for more information.
"""

try:
    from pymongo import ASCENDING, DESCENDING
except ImportError as e:  # pragma: no cover
    missing_dep_msg = (
        'The pymongo dependency is not installed. '
        'Install it with `pip install odata-v4-query[pymongo]` '
        'or install it directly with `pip install pymongo`.'
    )
    raise ImportError(missing_dep_msg) from e  # pragma: no cover

from typing import Any

from odata_v4_query.query_parser import ODataQueryOptions

from ._func import compute_skip_from_page
from .filter_parsers.mongo_filter_parser import MongoDBFilterNodeParser


class PyMongoQuery(dict):
    """Simple PyMongo query dictionary implementation.

    Supports the ``skip``, ``limit``, ``filter``, ``sort``
    and ``projection`` options.

    Examples
    --------
    >>> query = PyMongoQuery(
    ...     skip=1,
    ...     limit=10,
    ...     filter_={'name': 'John'},
    ...     sort=[('age', pymongo.ASCENDING)],
    ...     projection=['name', 'age'],
    ... )
    >>> query.skip
    1
    >>> query['skip']
    1
    >>> query.limit = 2
    >>> query.limit
    2
    >>> query['limit']
    2
    >>> query['limit'] = 3
    >>> query.limit
    3
    >>> db.users.find(
    ...     skip=query.skip,
    ...     limit=query.limit,
    ...     filter=query.filter,
    ...     sort=query.sort,
    ...     projection=query.projection,
    ... )
    >>> db.users.find(**query)

    """

    skip: int | None = None
    limit: int | None = None
    filter: dict[str, Any] | None = None
    sort: list[tuple[str, int]] | None = None
    projection: list[str] | None = None

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.skip = kwargs.get('skip')
        self.limit = kwargs.get('limit')
        self.filter = kwargs.get('filter')
        self.sort = kwargs.get('sort')
        self.projection = kwargs.get('projection')

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        super().__setitem__(name, value)

    def __setitem__(self, key: str, value: Any) -> None:
        super().__setitem__(key, value)
        super().__setattr__(key, value)


def get_query_from_options(
    options: ODataQueryOptions,
    search_fields: list[str] | None = None,
    parse_select: bool = False,
) -> PyMongoQuery:
    """Get a PyMongo query from OData query options.

    Parameters
    ----------
    options : ODataQueryOptions
        Parsed query options.
    search_fields : list[str] | None, optional
        Fields to search in if ``$search`` is used, by default None.
    parse_select : bool, optional
        If True, ``$select`` is parsed and applied as a projection,
        by default False.

    Returns
    -------
    PyMongoQuery
        PyMongo query.

    Notes
    -----
    #### Pagination
    If ``$page`` option is provided, ``$skip`` is overwritten
    using ``odata_v4_query.utils.compute_skip_from_page()``.
    If ``$top`` is not provided, it defaults to
    ``odata_v4_query.definitions.DEFAULT_LIMIT``.

    #### Search
    The ``$search`` option is only supported if ``search_fields``
    is provided.

    #### Select and projection
    The ``$select`` option is only supported if ``parse_select``
    is True.

    #### Unsupported options
    The ``$count``, ``$expand`` and ``$format`` options are
    not supported.

    #### Unsupported functions
    The following functions are not supported:
    - ``substring``
    - ``tolower``
    - ``toupper``

    Examples
    --------
    Usage:
    >>> from odata_v4_query import ODataQueryParser
    >>> from odata_v4_query.utils.pymongo import get_query_from_options
    >>> parser = ODataQueryParser()
    >>> options = parser.parse_query_string('$top=10&$skip=20')
    >>> query = get_query_from_options(options)

    Apply query to collection:
    >>> db.users.find(**query)
    >>> # or
    >>> db.users.find(
    ...     skip=query.skip,
    ...     limit=query.limit,
    ...     filter=query.filter,
    ...     sort=query.sort,
    ...     projection=query.projection,
    ... )

    Parsing ``$search`` option:
    >>> query = get_query_from_options(
    ...     options,
    ...     search_fields=['name', 'email']
    ... )

    Parsing ``$select`` option:
    >>> query = get_query_from_options(
    ...     options,
    ...     parse_select=True
    ... )

    """
    compute_skip_from_page(options)

    query = PyMongoQuery()

    if options.skip:
        query.skip = options.skip

    if options.top:
        query.limit = options.top

    if options.filter_:
        parser = MongoDBFilterNodeParser()
        filters = parser.parse(options.filter_)
        query.filter = filters

    if options.search and search_fields:
        search_filter = {
            '$or': [{field: {'$regex': options.search}} for field in search_fields],
        }
        if query.filter:
            query.filter = {'$and': [query.filter, search_filter]}
        else:
            query.filter = search_filter

    if options.orderby:
        sort_args = []
        for item in options.orderby:
            direction = DESCENDING if item.direction == 'desc' else ASCENDING
            sort_args.append((item.field, direction))
        query.sort = sort_args

    if options.select and parse_select:
        query.projection = options.select

    return query



"""Base class for filter node parsers."""

from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from odata_v4_query.definitions import (
    AND,
    COMPARISON_OPERATORS,
    CONTAINS,
    ENDSWITH,
    EQ,
    FUNCTION_ARITY,
    HAS,
    IN,
    LOGICAL_OPERATORS,
    NE,
    NIN,
    OR,
    STARTSWITH,
    SUBSTRING,
    TOLOWER,
    TOUPPER,
)
from odata_v4_query.errors import (
    UnexpectedEmptyArgumentsError,
    UnexpectedNullFiltersError,
    UnexpectedNullFunctionNameError,
    UnexpectedNullOperandError,
    UnexpectedNullOperatorError,
    UnexpectedNumberOfArgumentsError,
    UnexpectedTypeError,
    UnknownFunctionError,
    UnknownOperatorError,
)
from odata_v4_query.query_parser import FilterNode


@dataclass(frozen=True, kw_only=True)
class FunctionParser:
    func: Callable[..., FilterNode]
    arg_types: tuple[type, ...] | None = None


class BaseFilterNodeParser(ABC):
    """Base class for filter node parsers.

    The following methods must be implemented by subclasses:

    - **parse_startswith**: Parse a startswith function.
    - **parse_endswith**: Parse an endswith function.
    - **parse_contains**: Parse a contains function.
    - **parse_substring**: Parse a substring function.
    - **parse_tolower**: Parse a tolower function.
    - **parse_toupper**: Parse a toupper function.
    - **parse_membership_operators**: Parse an in/nin operator.
    - **parse_comparison_operators**: Parse an eq/ne/gt/ge/lt/le operator.
    - **parse_has_operator**: Parse a has operator.
    - **parse_and_or_operators**: Parse an and/or operator.
    - **parse_not_nor_operators**: Parse a not/nor operator.
    """

    def __init__(self) -> None:
        self._functions_map = {
            STARTSWITH: FunctionParser(func=self.parse_startswith, arg_types=(str,)),
            ENDSWITH: FunctionParser(func=self.parse_endswith, arg_types=(str,)),
            CONTAINS: FunctionParser(func=self.parse_contains, arg_types=(str,)),
            SUBSTRING: FunctionParser(
                func=self.parse_substring,
                arg_types=(int, int),
            ),
            TOLOWER: FunctionParser(func=self.parse_tolower),
            TOUPPER: FunctionParser(func=self.parse_toupper),
        }

    @abstractmethod
    def parse_startswith(self, field: str, value: Any) -> FilterNode:
        """Parse a startswith function.

        Parameters
        ----------
        field : str
            Field name.
        value : Any
            Value to compare.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        """
        ...

    @abstractmethod
    def parse_endswith(self, field: str, value: Any) -> FilterNode:
        """Parse an endswith function.

        Parameters
        ----------
        field : str
            Field name.
        value : Any
            Value to compare.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        """
        ...

    @abstractmethod
    def parse_contains(self, field: str, value: Any) -> FilterNode:
        """Parse a contains function.

        Parameters
        ----------
        field : str
            Field name.
        value : Any
            Value to compare.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        """
        ...

    @abstractmethod
    def parse_substring(self, field: str, start: int, length: int) -> FilterNode:
        """Parse a substring function.

        Parameters
        ----------
        field : str
            Field name.
        start : int
            Start index.
        length : int
            Length of the substring.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        """

    @abstractmethod
    def parse_tolower(self, field: str) -> FilterNode:
        """Parse a tolower function.

        Parameters
        ----------
        field : str
            Field name.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        """

    @abstractmethod
    def parse_toupper(self, field: str) -> FilterNode:
        """Parse a toupper function.

        Parameters
        ----------
        field : str
            Field name.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        """
        ...

    @abstractmethod
    def parse_membership_operators(
        self,
        left: Any,
        op_node: Any,
        right: Any,
    ) -> FilterNode:
        """Parse an in/nin operator.

        Parameters
        ----------
        left : Any
            Left operand.
        op_node : Any
            Operator node.
        right : Any
            Right operand.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        """
        ...

    @abstractmethod
    def parse_comparison_operators(
        self,
        left: Any,
        op_node: Any,
        right: Any,
    ) -> FilterNode:
        """Parse an eq/ne/gt/ge/lt/le operator.

        Parameters
        ----------
        left : Any
            Left operand.
        op_node : Any
            Operator node.
        right : Any
            Right operand.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        """
        ...

    @abstractmethod
    def parse_has_operator(self, left: Any, op_node: Any, right: Any) -> FilterNode:
        """Parse a has operator.

        Parameters
        ----------
        left : Any
            Left operand.
        op_node : Any
            Operator node.
        right : Any
            Right operand.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        """
        ...

    @abstractmethod
    def parse_and_or_operators(
        self,
        left: Any,
        op_node: Any,
        right: Any,
    ) -> FilterNode:
        """Parse an and/or operator.

        Parameters
        ----------
        left : Any
            Left operand.
        op_node : Any
            Operator node.
        right : Any
            Right operand.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        """
        ...

    @abstractmethod
    def parse_not_nor_operators(self, op_node: Any, right: Any) -> FilterNode:
        """Parse a not/nor operator.

        Parameters
        ----------
        op_node : Any
            Operator node.
        right : Any
            Right operand.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        """
        ...

    def parse(self, filter_node: FilterNode) -> Any:
        """Parse an AST to an ORM/ODM filter expression from the root node.

        Parameters
        ----------
        filter_node : FilterNode
            AST representing the parsed filter expression.

        Returns
        -------
        Any
            ORM/ODM filter expression.

        Raises
        ------
        UnexpectedNullFiltersError
            If the resulting filter is None.

        """
        filters = self.node_to_filter_expr(filter_node).value
        if filters is None:
            raise UnexpectedNullFiltersError(repr(filter_node))

        return filters

    def node_to_filter_expr(self, filter_node: FilterNode) -> FilterNode:
        """Recursively convert a filter node to an ORM/ODM filter expression.

        Parameters
        ----------
        filter_node : FilterNode
            AST representing the parsed filter expression.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        """
        if filter_node.type_ == 'function':
            return self.parse_function_node(filter_node)

        if filter_node.type_ == 'operator':
            left = None
            if filter_node.left is not None:
                left = self.node_to_filter_expr(filter_node.left)

            right = None
            if filter_node.right is not None:
                right = self.node_to_filter_expr(filter_node.right)

            return self.parse_operator_node(filter_node, left, right)

        return filter_node

    def parse_function_node(self, func_node: FilterNode) -> FilterNode:
        """Parse a function node to an ORM/ODM filter expression.

        Parameters
        ----------
        func_node : FilterNode
            AST representing the parsed filter expression.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        Raises
        ------
        UnexpectedNullFunctionNameError
            If function name is None.
        UnexpectedEmptyArgumentsError
            If arguments of the function are empty.
        UnexpectedNullOperandError
            If an operand is None.
        UnknownFunctionError
            If the function is unknown.
        UnexpectedNumberOfArgumentsError
            If the number of arguments does not match the expected number.

        """
        if not func_node.value:
            raise UnexpectedNullFunctionNameError(repr(func_node))

        if not func_node.arguments:
            raise UnexpectedEmptyArgumentsError(func_node.value)

        parser = self._functions_map.get(func_node.value)
        if parser is None:
            raise UnknownFunctionError(func_node.value)

        expected_args = FUNCTION_ARITY[func_node.value]
        if len(func_node.arguments) != expected_args:
            raise UnexpectedNumberOfArgumentsError(
                func_node.value,
                expected_args,
                len(func_node.arguments),
            )

        field = func_node.arguments[0].value
        if field is None:
            raise UnexpectedNullOperandError(func_node.value)

        return self._get_function_node(func_node.value, field, func_node.arguments[1:])

    def parse_operator_node(
        self,
        op_node: FilterNode,
        left: FilterNode | None,
        right: FilterNode | None,
    ) -> FilterNode:
        """Parse an operator node to an ORM/ODM filter expression.

        Parameters
        ----------
        op_node : FilterNode
            AST representing the parsed filter expression.
        left : FilterNode | None
            Left operand.
        right : FilterNode | None
            Right operand.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        Raises
        ------
        UnexpectedNullOperatorError
            If the operator is None.
        UnexpectedNullOperandError
            If an required operand is None.
        UnknownOperatorError
            If the operator is unknown.

        """
        if op_node.value is None:
            raise UnexpectedNullOperatorError(repr(op_node))

        if op_node.value in COMPARISON_OPERATORS:
            return self._parse_comparison_or_membership(op_node.value, left, right)

        if op_node.value == HAS:
            return self._parse_has(op_node.value, left, right)

        if op_node.value in LOGICAL_OPERATORS:
            return self._parse_logical(op_node.value, left, right)

        raise UnknownOperatorError(op_node.value)

    def _get_function_node(
        self,
        function_name: str,
        field: str,
        arguments: list[Any],
    ) -> FilterNode:
        """Get a function node.

        Parameters
        ----------
        function_name : str
            Function name.
        field : str
            Field name.
        arguments : list[Any]
            Arguments.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        Raises
        ------
        UnknownFunctionError
            If the function is unknown.
        UnexpectedNumberOfArgumentsError
            If the number of arguments does not match the expected number.
        UnexpectedNullOperandError
            If an operand is None.
        UnexpectedTypeError
            If an operand is of the wrong type.

        """
        parser = self._functions_map.get(function_name)
        if parser is None:  # pragma: no cover
            raise UnknownFunctionError(function_name)
        if not parser.arg_types:  # pragma: no cover
            return parser.func(field)
        if len(arguments) != len(parser.arg_types):  # pragma: no cover
            raise UnexpectedNumberOfArgumentsError(
                function_name,
                len(parser.arg_types),
                len(arguments),
            )
        args = []
        for i in range(len(parser.arg_types)):
            value = arguments[i].value
            if value is None:
                raise UnexpectedNullOperandError(function_name)
            type_ = parser.arg_types[i]
            try:
                value = type_(value)
            except ValueError as e:
                raise UnexpectedTypeError(function_name, type_.__name__, value) from e
            args.append(value)
        return parser.func(field, *args)

    def _parse_comparison_or_membership(
        self,
        operator: str,
        left: FilterNode | None,
        right: FilterNode | None,
    ) -> FilterNode:
        """Parse a comparison or membership operator.

        Parameters
        ----------
        operator : str
            Operator.
        left : FilterNode | None
            Left operand.
        right : FilterNode | None
            Right operand.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        Raises
        ------
        UnexpectedNullOperandError
            If an operand is None.

        """
        if left is None or right is None:
            raise UnexpectedNullOperandError(operator)

        if operator in (IN, NIN):
            if left.value is None or right.arguments is None:
                raise UnexpectedNullOperandError(operator)

            right.value = [arg.value for arg in right.arguments]
            return self.parse_membership_operators(
                left.value,
                operator,
                right.value,
            )

        if left.value is None or (operator not in (EQ, NE) and right.value is None):
            raise UnexpectedNullOperandError(operator)

        return self.parse_comparison_operators(
            left.value,
            operator,
            right.value,
        )

    def _parse_logical(
        self,
        operator: str,
        left: FilterNode | None,
        right: FilterNode | None,
    ) -> FilterNode:
        """Parse a logical operator.

        Parameters
        ----------
        operator : str
            Operator.
        left : FilterNode | None
            Left operand.
        right : FilterNode | None
            Right operand.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        Raises
        ------
        UnexpectedNullOperandError
            If an operand is None.

        """
        if operator in (AND, OR):
            if (
                left is None
                or right is None
                or left.value is None
                or right.value is None
            ):
                raise UnexpectedNullOperandError(operator)

            return self.parse_and_or_operators(
                left.value,
                operator,
                right.value,
            )

        if right is None or right.value is None:
            raise UnexpectedNullOperandError(operator)

        return self.parse_not_nor_operators(operator, right.value)

    def _parse_has(
        self,
        operator: str,
        left: FilterNode | None,
        right: FilterNode | None,
    ) -> FilterNode:
        """Parse a has operator.

        Parameters
        ----------
        operator : str
            Operator.
        left : FilterNode | None
            Left operand.
        right : FilterNode | None
            Right operand.

        Returns
        -------
        FilterNode
            New filter node containing the resulting ORM/ODM filter.

        Raises
        ------
        UnexpectedNullOperandError
            If an operand is None.

        """
        if left is None or right is None or left.value is None or right.value is None:
            raise UnexpectedNullOperandError(operator)

        return self.parse_has_operator(left.value, operator, right.value)

    def _get_value_filter_node(self, value: Any) -> FilterNode:
        """Get a filter node containing a value.

        Parameters
        ----------
        value : Any
            Value.

        Returns
        -------
        FilterNode
            New filter node containing the value.

        """
        return FilterNode(type_='value', value=value)
